from __future__ import annotations

from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import numpy as np

import gdsfactory as gf
from gdsfactory.component import Component, ComponentReference
from gdsfactory.port import Port, flipped
from gdsfactory.routing.get_route import get_route
from gdsfactory.typings import Route


def sort_key_west_to_east(port: Port) -> float:
    return port.x


def sort_key_east_to_west(port: Port) -> float:
    return -port.x


def sort_key_south_to_north(port: Port) -> float:
    return port.y


def sort_key_north_to_south(port: Port) -> float:
    return -port.y


def route_ports_to_side(
    ports: Union[Dict[str, Port], List[Port], Component, ComponentReference],
    side: str = "north",
    x: Optional[float] = None,
    y: Optional[float] = None,
    routing_func=get_route,
    **kwargs,
) -> Tuple[List[Route], List[Port]]:
    """Routes ports to a given side.

    Args:
        ports: list/dict/Component/ComponentReference to route to a side.
        side: 'north', 'south', 'east' or 'west'.
        x: position to route ports for east/west. None, uses most east/west value.
        y: position to route ports for south/north. None, uses most north/south value.
        routing_func: the routing function. By default uses `get_route`.

    Keyword Args:
      radius: in um.
      separation: in um.
      extend_bottom/extend_top for east/west routing.
      extend_left, extend_right for south/north routing.

    Returns:
        List of routes: with routing elements.
        List of ports: of the new ports.

    .. plot::
        :include-source:

        import gdsfactory as gf

        c = gf.Component('sample_route_sides')
        dummy = gf.components.nxn(north=2, south=2, west=2, east=2)
        sides = ["north", "south", "east", "west"]
        d = 100
        positions = [(0, 0), (d, 0), (d, d), (0, d)]

        for pos, side in zip(positions, sides):
            dummy_ref = dummy.ref(position=pos)
            c.add(dummy_ref)
            routes, ports = gf.routing.route_ports_to_side(dummy_ref, side, layer=(1, 0))
            for route in routes:
                c.add(route.references)
            for i, p in enumerate(ports):
                c.add_port(name=f"{side[0]}{i}", port=p)

        c.plot()

    """
    if not ports:
        return [], []

    # Accept list of ports, Component or dict of ports
    if isinstance(ports, dict):
        ports = list(ports.values())

    elif isinstance(ports, (Component, ComponentReference)):
        ports = list(ports.ports.values())

    # Choose which
    if side in {"north", "south"}:
        func_route = route_ports_to_y
        xy = y if y is not None else side
    elif side in {"west", "east"}:
        xy = x if x is not None else side
        func_route = route_ports_to_x
    else:
        raise ValueError(f"side = {side} not valid (north, south, west, east)")

    return func_route(ports, xy, routing_func=routing_func, **kwargs)


def route_ports_to_north(list_ports, **kwargs):
    return route_ports_to_side(list_ports, side="north", **kwargs)


def route_ports_to_south(list_ports, **kwargs):
    return route_ports_to_side(list_ports, side="south", **kwargs)


def route_ports_to_west(list_ports, **kwargs):
    return route_ports_to_side(list_ports, side="west", **kwargs)


def route_ports_to_east(list_ports, **kwargs):
    return route_ports_to_side(list_ports, side="east", **kwargs)


def route_ports_to_x(
    list_ports: List[Port],
    x: Union[float, str] = "east",
    separation: float = 10.0,
    radius: float = 10.0,
    extend_bottom: float = 0.0,
    extend_top: float = 0.0,
    extension_length: float = 0.0,
    y0_bottom: Optional[float] = None,
    y0_top: Optional[float] = None,
    routing_func: Callable = get_route,
    backward_port_side_split_index: int = 0,
    start_straight_length: float = 0.01,
    dx_start: Optional[float] = None,
    dy_start: Optional[float] = None,
    **routing_func_args,
) -> Tuple[List[Route], List[Port]]:
    """Returns route to x.

    Args:
        list_ports: reasonably well behaved list of ports.
           ports facing north ports are norther than any other ports
           ports facing south ports are souther ...
           ports facing west ports are the wester ...
           ports facing east ports are the easter ...
        x: float or string.
           if float: x coordinate to which the ports will be routed
           if string: "east" -> route to east
           if string: "west" -> route to west
        separation: in um.
        radius: in um.
        extend_bottom: in um.
        extend_top: in um.
        extension_length: in um.
        y0_bottom: in um.
        y0_top: in um.
        routing_func: to route.
        backward_port_side_split_index: integer represents and index in the list of backwards ports (bottom to top)
                all ports with an index strictly lower or equal are routed bottom
            all ports with an index larger or equal are routed top
        dx_start: override minimum starting x distance.
        dy_start: override minimum starting y distance.

    Returns:
        routes: list of routes
        ports: list of the new optical ports

    1. routes the bottom-half of the ports facing opposite side of x
    2. routes the south ports
    3. front ports
    4. north ports

    """
    north_ports = [p for p in list_ports if p.orientation == 90]
    south_ports = [p for p in list_ports if p.orientation == 270]
    east_ports = [p for p in list_ports if p.orientation == 0]
    west_ports = [p for p in list_ports if p.orientation == 180]

    epsilon = 1.0
    a = epsilon + max(radius, separation)
    bx = epsilon + max(radius, dx_start) if dx_start else a
    by = epsilon + max(radius, dy_start) if dy_start else a

    xs = [p.x for p in list_ports]
    ys = [p.y for p in list_ports]

    if y0_bottom is None:
        y0_bottom = min(ys) - by

    y0_bottom -= extend_bottom

    if y0_top is None:
        y0_top = max(ys) + (max(radius, dy_start) if dy_start else a)
    y0_top += extend_top

    if x == "west" and extension_length > 0:
        extension_length = -extension_length

    if x == "east":
        x = max(p.x for p in list_ports) + bx
    elif x == "west":
        x = min(p.x for p in list_ports) - bx
    elif isinstance(x, (float, int)):
        pass
    else:
        raise ValueError(f"x={x!r} should be a float or east or west")

    if x < min(xs):
        sort_key_north = sort_key_west_to_east
        sort_key_south = sort_key_west_to_east
        forward_ports = west_ports
        backward_ports = east_ports
        angle = 0

    elif x > max(xs):
        sort_key_south = sort_key_east_to_west
        sort_key_north = sort_key_east_to_west
        forward_ports = east_ports
        backward_ports = west_ports
        angle = 180
    else:
        raise ValueError("x should be either to the east or to the west of all ports")

    # forward_ports.sort()
    north_ports.sort(key=sort_key_north)
    south_ports.sort(key=sort_key_south)
    forward_ports.sort(key=sort_key_south_to_north)

    backward_ports.sort(key=sort_key_south_to_north)
    backward_ports_thru_south = backward_ports[:backward_port_side_split_index]
    backward_ports_thru_north = backward_ports[backward_port_side_split_index:]
    backward_ports_thru_south.sort(key=sort_key_south_to_north)
    backward_ports_thru_north.sort(key=sort_key_north_to_south)

    routes = []
    ports = []

    def add_port(
        p, y, l_elements, l_ports, start_straight_length=start_straight_length
    ) -> None:
        new_port = p.copy(name=f"{p.name}_new")
        new_port.orientation = angle
        new_port.center = (x + extension_length, y)
        l_elements += [
            routing_func(
                p,
                new_port,
                start_straight_length=start_straight_length,
                radius=radius,
                **routing_func_args,
            )
        ]
        l_ports += [new_port.flip()]

    y_optical_bot = y0_bottom
    for p in south_ports:
        add_port(p, y_optical_bot, routes, ports)
        y_optical_bot -= separation

    for p in forward_ports:
        add_port(p, p.y, routes, ports)

    y_optical_top = y0_top
    for p in north_ports:
        add_port(p, y_optical_top, routes, ports)
        y_optical_top += separation

    start_straight_length_section = start_straight_length
    max_x = max(xs)
    min_x = min(xs)

    for p in backward_ports_thru_north:
        # Extend ports if necessary
        if angle == 0 and p.x < max_x:
            start_straight_length_section = max_x - p.x
        elif angle == 180 and p.x > min_x:
            start_straight_length_section = p.x - min_x
        else:
            start_straight_length_section = 0

        add_port(
            p,
            y_optical_top,
            routes,
            ports,
            start_straight_length=start_straight_length + start_straight_length_section,
        )
        y_optical_top += separation
        start_straight_length += separation

    start_straight_length_section = start_straight_length
    for p in backward_ports_thru_south:
        # Extend ports if necessary
        if angle == 0 and p.x < max_x:
            start_straight_length_section = max_x - p.x
        elif angle == 180 and p.x > min_x:
            start_straight_length_section = p.x - min_x
        else:
            start_straight_length_section = 0

        add_port(
            p,
            y_optical_bot,
            routes,
            ports,
            start_straight_length=start_straight_length + start_straight_length_section,
        )
        y_optical_bot -= separation
        start_straight_length += separation

    return routes, ports


def route_ports_to_y(
    list_ports: List[Port],
    y: Union[float, str] = "north",
    separation: float = 10.0,
    radius: float = 10.0,
    x0_left: Optional[float] = None,
    x0_right: Optional[float] = None,
    extension_length: float = 0.0,
    extend_left: float = 0.0,
    extend_right: float = 0.0,
    routing_func: Callable = get_route,
    backward_port_side_split_index: int = 0,
    start_straight_length: float = 0.01,
    dx_start: float = None,
    dy_start: float = None,
    **routing_func_args: Dict[Any, Any],
) -> Tuple[List[Route], List[Port]]:
    """Args are the following.

        list_ports: reasonably well behaved list of ports.
           ports facing north ports are norther than any other ports
           ports facing south ports are souther ...
           ports facing west ports are the wester ...
           ports facing east ports are the easter ...

        y: float or string.
               if float: y coordinate to which the ports will be routed
               if string: "north" -> route to north
               if string: "south" -> route to south

        backward_port_side_split_index: integer
               this integer represents and index in the list of backwards ports
                   (sorted from left to right)
               all ports with an index strictly larger are routed right
               all ports with an index lower or equal are routed left
        separation: in um.
        radius: in um.

    Returns:
        - a list of Routes
        - a list of the new optical ports

    First route the bottom-half of the back ports
        (back ports are the one facing opposite side of x)
    Then route the south ports
    then the front ports
    then the north ports
    """
    if y == "south" and extension_length > 0:
        extension_length = -extension_length

    da = 45
    north_ports = [
        p for p in list_ports if p.orientation > 90 - da and p.orientation < 90 + da
    ]
    south_ports = [
        p for p in list_ports if p.orientation > 270 - da and p.orientation < 270 + da
    ]
    east_ports = [
        p for p in list_ports if p.orientation < da or p.orientation > 360 - da
    ]
    west_ports = [
        p for p in list_ports if p.orientation < 180 + da and p.orientation > 180 - da
    ]

    epsilon = 1.0
    a = radius + max(radius, separation)
    bx = epsilon + max(radius, dx_start) if dx_start else a
    by = epsilon + max(radius, dy_start) if dy_start else a

    xs = [p.x for p in list_ports]
    ys = [p.y for p in list_ports]

    if x0_left is None:
        x0_left = min(xs) - bx
    x0_left -= extend_left

    if x0_right is None:
        x0_right = max(xs) + (max(radius, dx_start) if dx_start else a)
    x0_right += extend_right

    if y == "north":
        y = (
            max(
                p.y + a * np.abs(np.cos(p.orientation * np.pi / 180))
                for p in list_ports
            )
            + by
        )
    elif y == "south":
        y = (
            min(
                p.y - a * np.abs(np.cos(p.orientation * np.pi / 180))
                for p in list_ports
            )
            - by
        )
    elif isinstance(y, (float, int)):
        pass
    if y <= min(ys):
        sort_key_east = sort_key_south_to_north
        sort_key_west = sort_key_south_to_north
        forward_ports = south_ports
        backward_ports = north_ports
        angle = 90.0

    elif y >= max(ys):
        sort_key_west = sort_key_north_to_south
        sort_key_east = sort_key_north_to_south
        forward_ports = north_ports
        backward_ports = south_ports
        angle = -90.0
    else:
        raise ValueError("y should be either to the north or to the south of all ports")

    west_ports.sort(key=sort_key_west)
    east_ports.sort(key=sort_key_east)
    forward_ports.sort(key=sort_key_west_to_east)
    backward_ports.sort(key=sort_key_east_to_west)

    backward_ports.sort(key=sort_key_west_to_east)
    backward_ports_thru_west = backward_ports[:backward_port_side_split_index]
    backward_ports_thru_east = backward_ports[backward_port_side_split_index:]

    backward_ports_thru_west.sort(key=sort_key_west_to_east)
    backward_ports_thru_east.sort(key=sort_key_east_to_west)

    routes = []
    ports = []

    def add_port(
        p, x, l_elements, l_ports, start_straight_length=start_straight_length
    ):
        new_port = p.copy()
        new_port.orientation = angle
        new_port.center = (x, y + extension_length)

        if np.sum(np.abs((new_port.center - p.center) ** 2)) < 1e-12:
            l_ports += [flipped(new_port)]
            return

        try:
            l_elements += [
                routing_func(
                    p,
                    new_port,
                    start_straight_length=start_straight_length,
                    radius=radius,
                    **routing_func_args,
                )
            ]
            l_ports += [flipped(new_port)]

        except Exception as error:
            raise ValueError(
                f"Could not connect {p.name!r} to {new_port.name!r}"
            ) from error

    x_optical_left = x0_left
    for p in west_ports:
        add_port(p, x_optical_left, routes, ports)
        x_optical_left -= separation

    for p in forward_ports:
        add_port(p, p.x, routes, ports)

    x_optical_right = x0_right
    for p in east_ports:
        add_port(p, x_optical_right, routes, ports)
        x_optical_right += separation

    start_straight_length_section = start_straight_length
    for p in backward_ports_thru_east:
        add_port(
            p,
            x_optical_right,
            routes,
            ports,
            start_straight_length=start_straight_length_section,
        )
        x_optical_right += separation
        start_straight_length_section += separation

    start_straight_length_section = start_straight_length
    for p in backward_ports_thru_west:
        add_port(
            p,
            x_optical_left,
            routes,
            ports,
            start_straight_length=start_straight_length_section,
        )
        x_optical_left -= separation
        start_straight_length_section += separation

    return routes, ports


@gf.cell
def _sample_route_side() -> Component:
    c = Component()
    xs = [0.0, 10.0, 25.0, 50.0]
    ys = [0.0, 10.0, 25.0, 50.0]
    a = 5
    xl = min(xs) - a
    xr = max(xs) + a
    yb = min(ys) - a
    yt = max(ys) + a

    layer = (1, 0)

    c.add_polygon([(xl, yb), (xl, yt), (xr, yt), (xr, yb)], layer)

    for i, y in enumerate(ys):
        p0 = (xl, y)
        p1 = (xr, y)
        c.add_port(name=f"W{i}", center=p0, orientation=180, width=0.5, layer=layer)
        c.add_port(name=f"E{i}", center=p1, orientation=0, width=0.5, layer=layer)

    for i, x in enumerate(xs):
        p0 = (x, yb)
        p1 = (x, yt)
        c.add_port(name=f"S{i}", center=p0, orientation=270, width=0.5, layer=layer)
        c.add_port(name=f"N{i}", center=p1, orientation=90, width=0.5, layer=layer)

    return c


@gf.cell
def _sample_route_sides() -> Component:
    c = Component()
    dummy = _sample_route_side()
    sides = ["north", "south", "east", "west"]
    positions = [(0, 0), (400, 0), (400, 400), (0, 400)]
    for pos, side in zip(positions, sides):
        dummy_ref = dummy.ref(position=pos)
        c.add(dummy_ref)
        routes, ports = route_ports_to_side(dummy_ref, side, layer=(1, 0))
        for route in routes:
            c.add(route.references)
        for i, p in enumerate(ports):
            c.add_port(name=f"{side[0]}{i}", port=p)
    return c


if __name__ == "__main__":
    c = Component("sample_route_sides")
    dummy = gf.components.nxn(north=2, south=2, west=2, east=2)
    sides = ["north", "south", "east", "west"]
    d = 100
    positions = [(0, 0), (d, 0), (d, d), (0, d)]

    for pos, side in zip(positions, sides):
        dummy_ref = dummy.ref(position=pos)
        c.add(dummy_ref)
        routes, ports = route_ports_to_side(dummy_ref, side, layer=(1, 0))
        for route in routes:
            c.add(route.references)
        for i, p in enumerate(ports):
            c.add_port(name=f"{side[0]}{i}", port=p)

    # c.plot()
    c.show(show_ports=True)



from __future__ import annotations

from typing import Tuple

import numpy as np

from gdsfactory.cell import cell
from gdsfactory.component import Component
from gdsfactory.components.rectangle import rectangle
from gdsfactory.geometry.offset import offset
from gdsfactory.typings import LayerSpec


@cell
def resistance_meander(
    pad_size: Tuple[float, float] = (50.0, 50.0),
    num_squares: int = 1000,
    width: float = 1.0,
    res_layer: LayerSpec = "MTOP",
    pad_layer: LayerSpec = "MTOP",
    gnd_layer: LayerSpec = "MTOP",
) -> Component:
    """Return meander to test resistance.

    based on phidl.geometry

    Args:
        pad_size: Size of the two matched impedance pads (microns).
        num_squares: Number of squares comprising the resonator wire.
        width: The width of the squares (microns).
        res_layer: resistance layer.
        pad_layer: pad layer.
        gnd_layer: ground layer.
    """
    x = pad_size[0]
    z = pad_size[1]

    # Checking validity of input
    if x <= 0 or z <= 0:
        raise ValueError("Pad must have positive, real dimensions")
    elif width > z:
        raise ValueError("Width of cell cannot be greater than height of pad")
    elif num_squares <= 0:
        raise ValueError("Number of squares must be a positive real number")
    elif width <= 0:
        raise ValueError("Width of cell must be a positive real number")

    # Performing preliminary calculations
    num_rows = int(np.floor(z / (2 * width)))
    if num_rows % 2 == 0:
        num_rows -= 1
    num_columns = num_rows - 1
    squares_in_row = (num_squares - num_columns - 2) / num_rows

    # Compensating for weird edge cases
    if squares_in_row < 1:
        num_rows = round(num_rows / 2) - 2
        squares_in_row = 1
    if width * 2 > z:
        num_rows = 1
        squares_in_row = num_squares - 2

    length_row = squares_in_row * width

    # Creating row/column corner combination structure
    T = Component()
    Row = rectangle(size=(length_row, width), layer=res_layer)
    Col = rectangle(size=(width, width), layer=res_layer)

    T.add_ref(Row)
    col = T.add_ref(Col)
    col.move([length_row - width, -width])

    # Creating entire straight net
    N = Component("net")
    n = 1
    for i in range(num_rows):
        d = N.add_ref(T) if i != num_rows - 1 else N.add_ref(Row)
        if n % 2 == 0:
            d.mirror(p1=(d.x, d.ymax), p2=(d.x, d.ymin))
        d.movey(-(n - 1) * T.ysize)
        n += 1
    d = N.add_ref(Col).movex(-width)
    d = N.add_ref(Col).move([length_row, -(n - 2) * T.ysize])

    # Creating pads
    P = Component()
    pad1 = rectangle(size=(x, z), layer=pad_layer)
    gnd1 = gnd2 = offset(pad1, distance=-5, layer=gnd_layer)
    pad1_ref = P.add_ref(pad1)
    pad1_ref.movex(-x - width)
    pad2_ref = P.add_ref(pad1)
    pad2_ref.movex(length_row + width)
    gnd1_ref = P.add_ref(gnd1)
    gnd1_ref.center = pad1_ref.center
    gnd2_ref = P.add_ref(gnd2)
    net = P.add_ref(N)
    net.y = pad1_ref.y
    gnd2_ref.center = pad2_ref.center
    gnd2_ref.movex(2.5)
    P.absorb(net)
    P.absorb(gnd1_ref)
    P.absorb(gnd2_ref)
    P.absorb(pad1_ref)
    P.absorb(pad2_ref)
    return P


if __name__ == "__main__":
    c = resistance_meander(res_layer="M1", width=0.5)
    c.show(show_ports=True)



from __future__ import annotations

import numpy as np
from SiPANN.scee import Waveguide

import gdsfactory as gf


def bend_euler(
    radius: float = 10.0,
    width: float = 0.5,
    thickness: float = 0.22,
    angle: float = 90,
    sw_angle: float = 90.0,
    **kwargs,
):
    """Return simphony Model for a bend using a straight.

    FIXME. this is fake bend! need to create a better model

    Args:
        radius: Radius of straight in microns.
        width: Width of the straights in microns.
        thickness: Thickness of the straights in microns.
        angle: Number of deg of circle that bent straight transverses.
        sw_angle: Sidewall angle from horizontal in degrees, ie 90 makes a square. Defaults to 90.
        kwargs: geometrical args that this model ignores.

    """
    c = gf.c.bend_euler(radius=radius, **kwargs)
    length = c.info["length"] * 1e3
    angle = np.deg2rad(angle)
    width *= 1e3
    thickness *= 1e3

    return Waveguide(width=width, thickness=thickness, sw_angle=sw_angle, length=length)


if __name__ == "__main__":
    import matplotlib.pyplot as plt

    c = bend_euler()
    wavelength = np.linspace(1500, 1600, 500)
    t = c.predict(wavelength)

    plt.figure(figsize=(15, 5))
    plt.subplot(121)
    plt.plot(wavelength, np.angle(t), label="t")
    plt.xlabel("Wavelength (nm)")
    plt.ylabel("Phase (rad)")
    plt.title(r"Transmission at $\lambda \approx 1550nm$")
    plt.legend()
    plt.show()



from __future__ import annotations

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.pad import pad as pad_function
from gdsfactory.components.rectangle import rectangle
from gdsfactory.typings import ComponentSpec, LayerSpec


@gf.cell
def pads_shorted(
    pad: ComponentSpec = pad_function,
    columns: int = 8,
    pad_spacing: float = 150.0,
    layer_metal: LayerSpec = "M3",
    metal_width: float = 10,
) -> Component:
    """Returns a 1D array of shorted_pads.

    Args:
        pad: pad spec.
        columns: number of columns.
        pad_spacing: in um
        layer_metal: for the short.
        metal_width: for the short.
    """
    c = Component()
    pad = gf.get_component(pad)
    for i in range(columns):
        pad_ref = c.add_ref(pad)
        pad_ref.movex(i * pad_spacing - columns / 2 * pad_spacing + pad_spacing / 2)

    short = rectangle(
        size=(pad_spacing * (columns - 1), metal_width),
        layer=layer_metal,
        centered=True,
    )
    c.add_ref(short)
    return c


if __name__ == "__main__":
    c = pads_shorted(metal_width=20)
    c.show(show_ports=True)



"""Lets create a crossing component with two references to other components (crossing_arm).

- add references to a component (one of the arm references rotated)
- add ports from the child references into the parent cell
- use Component.auto_rename_ports() to rename ports according to their location

"""

from __future__ import annotations

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.typings import Layer


@gf.cell
def test_crossing_arm(
    wg_width: float = 0.5,
    r1: float = 3.0,
    r2: float = 1.1,
    taper_width: float = 1.2,
    taper_length: float = 3.4,
    layer_wg: Layer = (1, 0),
    layer_slab: Layer = (2, 0),
) -> Component:
    """Crossing arm."""
    c = gf.Component()
    c << gf.components.ellipse(radii=(r1, r2), layer=layer_slab)

    xmax = taper_length + taper_width / 2
    h = wg_width / 2
    taper_points = [
        (-xmax, h),
        (-taper_width / 2, taper_width / 2),
        (taper_width / 2, taper_width / 2),
        (xmax, h),
        (xmax, -h),
        (taper_width / 2, -taper_width / 2),
        (-taper_width / 2, -taper_width / 2),
        (-xmax, -h),
    ]

    c.add_polygon(taper_points, layer=layer_wg)

    c.add_port(
        name="o1", center=(-xmax, 0), orientation=180, width=wg_width, layer=layer_wg
    )

    c.add_port(
        name="o2", center=(xmax, 0), orientation=0, width=wg_width, layer=layer_wg
    )
    return c


@gf.cell  # This decorator will generate a good name for the component
def test_crossing() -> Component:
    c = gf.Component()
    arm = test_crossing_arm()

    # Create two arm references. One has a 90Deg rotation
    arm_h = arm.ref(position=(0, 0))
    arm_v = arm.ref(position=(0, 0), rotation=90)

    # Add each arm to the component
    # Also add the ports
    port_id = 0
    for a in [arm_h, arm_v]:
        c.add(a)
        for p in a.ports.values():
            # Here we don't care too much about the name we give to the ports
            # since they will be renamed. We just want the names to be unique
            c.add_port(name=f"{port_id}", port=p)
            port_id += 1

    c.auto_rename_ports()
    return c


if __name__ == "__main__":
    c = test_crossing()
    c.show(show_ports=True)



from __future__ import annotations

from simphony.libraries import siepic

from gdsfactory.simulation.simphony.components.mmi1x2 import mmi1x2


def mzi(L0=1, DL=100.0, L2=10.0, y_model_factory=mmi1x2, wg=siepic.Waveguide):
    """Mzi circuit model.

    Args:
        L0 (um): vertical length for both and top arms
        DL (um): bottom arm extra length, delta_length = 2*DL
        L2 (um): L_top horizontal length

    Return: mzi circuit model

    .. code::

               __L2__
               |      |
               L0     L0r
               |      |
     splitter==|      |==recombiner
               |      |
               L0     L0r
               |      |
               DL     DL
               |      |
               |__L2__|


    .. plot::
      :include-source:

      import gdsfactory as gf

      c = gf.c.mzi(L0=0.1, DL=0, L2=10)
      gf.plotgds(c)


    .. plot::
        :include-source:

        import gdsfactory.simulation.simphony as gs
        import gdsfactory.simulation.simphony.components as gc

        c = gc.mzi()
        gs.plot_circuit(c)

    """
    y_splitter = y_model_factory() if callable(y_model_factory) else y_model_factory
    y_recombiner = y_model_factory() if callable(y_model_factory) else y_model_factory
    wg_long = wg(length=(2 * L0 + 2 * DL + L2) * 1e-6)
    wg_short = wg(length=(2 * L0 + L2) * 1e-6)

    y_recombiner.pins[0].rename("o2")

    y_splitter[1].connect(wg_long)
    y_splitter[2].connect(wg_short)
    y_recombiner.multiconnect(None, wg_long, wg_short)

    return y_splitter.circuit.to_subcircuit("mzi")


if __name__ == "__main__":
    import matplotlib.pyplot as plt

    from gdsfactory.simulation.simphony import plot_circuit

    c = mzi()
    plot_circuit(c)
    plt.show()



from __future__ import annotations

import itertools
import pathlib
from pathlib import Path
from typing import Union

import gdstk

from gdsfactory.component import Component
from gdsfactory.read.import_gds import import_gds

COUNTER = itertools.count()


def xor_polygons(A: Component, B: Component, hash_geometry: bool = True):
    """Given two devices A and B, performs a layer-by-layer XOR diff between A \
    and B, and returns polygons representing the differences between A and B.

    Adapted from lytest/kdb_xor.py

    """
    # first do a geometry hash to vastly speed up if they are equal
    if hash_geometry and (A.hash_geometry() == B.hash_geometry()):
        return Component()

    D = Component()
    A_polys = A.get_polygons(by_spec=True)
    B_polys = B.get_polygons(by_spec=True)
    A_layers = A_polys.keys()
    B_layers = B_polys.keys()
    all_layers = set()
    all_layers.update(A_layers)
    all_layers.update(B_layers)
    for layer in all_layers:
        if (layer in A_layers) and (layer in B_layers):
            p = gdstk.boolean(
                A_polys[layer],
                B_polys[layer],
                operation="xor",
                precision=0.001,
                layer=layer[0],
                datatype=layer[1],
            )
        elif layer in A_layers:
            p = A_polys[layer]
        elif layer in B_layers:
            p = B_polys[layer]
        if p is not None:
            for polygon in p:
                D.add_polygon(polygon, layer=layer)
    return D


def gdsdiff(
    component1: Union[Path, Component, str],
    component2: Union[Path, Component, str],
    name: str = "TOP",
    xor: bool = True,
) -> Component:
    """Compare two Components.

    Args:
        component1: Component or path to gds file (reference).
        component2: Component or path to gds file (run).
        name: name of the top cell.
        xor: makes boolean operation.

    Returns:
        Component with both cells (xor, common and diffs).
    """
    if isinstance(component1, (str, pathlib.Path)):
        component1 = import_gds(str(component1), flatten=True, name=f"{name}_old")
    if isinstance(component2, (str, pathlib.Path)):
        component2 = import_gds(str(component2), flatten=True, name=f"{name}_new")

    component1 = component1.copy()
    component2 = component2.copy()

    component1.name = f"{name}_old"
    component2.name = f"{name}_new"

    top = Component(name=f"{name}_diffs")
    ref1 = top << component1
    ref2 = top << component2

    if xor:
        diff = xor_polygons(ref1, ref2, hash_geometry=False)
        diff.name = f"{name}_xor"
        top.add_ref(diff)

    return top


if __name__ == "__main__":
    import sys

    if len(sys.argv) != 3:
        print("Usage: gdsdiff <mask_v1.gds> <mask_v2.gds>")
        print("Note that you need to have KLayout opened with klive running")
        sys.exit()
    c = gdsdiff(sys.argv[1], sys.argv[2])
    c.show(show_ports=True)



from __future__ import annotations

import pathlib

from jsondiff import diff
from omegaconf import OmegaConf

import gdsfactory as gf
import gdsfactory.simulation.gtidy3d as gt

# from gdsfactory.simulation.gtidy3d.get_results import get_sim_hash
# def test_simulation_hash() -> None:
#     component = gf.components.straight(length=3)
#     sim = gt.get_simulation(component=component)
#     sim_hash = get_sim_hash(sim)
#     sim_hash_reference = "4b0cd0b69d0e1c32a1bf30e1f1ea5b27"
#     assert sim_hash == sim_hash_reference, f"sim_hash_reference = {sim_hash!r}"


def test_simulation(overwrite: bool = False) -> None:
    """export sim in JSON, and then load it again."""
    component = gf.components.straight(length=3)
    sim = gt.get_simulation(component=component)

    if overwrite:
        sim.to_file("sim_ref.yaml")  # uncomment to overwrite material

    sim.to_file("sim_run.yaml")

    dirpath = pathlib.Path(__file__).parent
    dref = OmegaConf.load(dirpath / "sim_ref.yaml")
    drun = OmegaConf.load(dirpath / "sim_run.yaml")

    d = diff(dref, drun)
    assert len(d) == 0, d


if __name__ == "__main__":
    # test_simulation_hash()
    test_simulation(overwrite=True)

    # test_simulation()
    # component = gf.components.straight(length=3)
    # sim = gt.get_simulation(component=component)
    # sim.to_file("sim_ref.yaml")
    # sim.to_file("sim_run.yaml")

    # dref = OmegaConf.load("sim_ref.yaml")
    # drun = OmegaConf.load("sim_run.yaml")

    # d = diff(dref, drun)



"""Straight Ge photodetector."""
from __future__ import annotations

from functools import partial
from typing import Optional

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.taper import taper as taper_func
from gdsfactory.components.via_stack import via_stack_slab_m3
from gdsfactory.cross_section import pn_ge_detector_si_contacts
from gdsfactory.typings import ComponentSpec, CrossSectionSpec

default_taper = partial(
    taper_func, length=20.0, width1=0.5, width2=0.8, cross_section="strip"
)


@gf.cell
def ge_detector_straight_si_contacts(
    length: float = 80.0,
    cross_section: CrossSectionSpec = pn_ge_detector_si_contacts,
    via_stack: ComponentSpec = via_stack_slab_m3,
    via_stack_width: float = 10.0,
    via_stack_spacing: float = 5.0,
    taper: Optional[ComponentSpec] = default_taper,
    **kwargs,
) -> Component:
    """Returns a straight Ge on Si detector with silicon contacts.

    There are no contacts on the Ge. These detectors could have lower
    dark current and sensitivity compared to those with contacts in the
    Ge. See Chen et al., "High-Responsivity Low-Voltage 28-Gb/s Ge p-i-n
    Photodetector With Silicon Contacts", Journal of Lightwave Technology 33(4), 2015.

    https://doi.org/10.1109/JLT.2014.2367134

    Args:
        length: total length of the waveguide including the tapers.
        cross_section: for the waveguide.
        via_stack: for the via_stacks.
        via_stack_width: width of the via_stack.
        via_stack_spacing: spacing between via_stacks.
        taper: optional taper to transition from the input waveguide
            into the absorption region.
        kwargs: cross_section settings.
    """
    c = Component()
    if taper:
        taper = gf.get_component(taper)
        length -= 2 * taper.get_ports_xsize()

    wg = c << gf.components.straight(
        cross_section=cross_section,
        length=length,
        **kwargs,
    )

    if taper:
        t1 = c << taper
        t1.connect("o2", wg.ports["o1"])
        c.add_port("o1", port=t1.ports["o1"])

    else:
        c.add_ports(wg.get_ports_list())

    via_stack_length = length
    via_stack_top = c << via_stack(
        size=(via_stack_length, via_stack_width),
    )
    via_stack_bot = c << via_stack(
        size=(via_stack_length, via_stack_width),
    )

    via_stack_bot.xmin = wg.xmin
    via_stack_top.xmin = wg.xmin

    via_stack_top.ymin = +via_stack_spacing / 2
    via_stack_bot.ymax = -via_stack_spacing / 2

    c.add_ports(via_stack_bot.ports, prefix="bot_")
    c.add_ports(via_stack_top.ports, prefix="top_")
    return c


if __name__ == "__main__":
    comp = ge_detector_straight_si_contacts()
    # print(c.ports.keys())
    comp.show(show_ports=True)



from __future__ import annotations

from gdsfactory.simulation.modes.find_neff_ng_dw_dh import find_neff_ng_dw_dh


def test_dw_dh(dataframe_regression) -> None:
    df = find_neff_ng_dw_dh(steps=1, resolution=10, cache=None)
    if dataframe_regression:
        dataframe_regression.check(df, default_tolerance=dict(atol=1e-2, rtol=1e-2))
    else:
        print(df)


if __name__ == "__main__":
    test_dw_dh(None)



"""Based on phidl.geometry."""

from __future__ import annotations

import gdstk

import gdsfactory as gf
from gdsfactory.component_layout import Polygon, _parse_layer
from gdsfactory.typings import Component, ComponentReference, LayerSpec


@gf.cell
def offset(
    elements: Component,
    distance: float = 0.1,
    use_union: bool = True,
    precision: float = 1e-4,
    join: str = "miter",
    tolerance: int = 2,
    layer: LayerSpec = "WG",
) -> Component:
    """Returns an element containing all polygons with an offset Shrinks or \
    expands a polygon or set of polygons.

    Args:
        elements: Component(/Reference), list of Component(/Reference), or Polygon
          Polygons to offset or Component containing polygons to offset.
        distance: Distance to offset polygons. Positive values expand, negative shrink.
        precision: Desired precision for rounding vertex coordinates.
        num_divisions: The number of divisions with which the geometry is divided into
          multiple rectangular regions. This allows for each region to be
          processed sequentially, which is more computationally efficient.
        join: {'miter', 'bevel', 'round'} Type of join used to create polygon offset
        tolerance: For miter joints, this number must be at least 2 represents the
          maximal distance in multiples of offset between new vertices and their
          original position before beveling to avoid spikes at acute joints. For
          round joints, it indicates the curvature resolution in number of
          points per full circle.
        layer: Specific layer to put polygon geometry on.

    Returns:
        Component containing a polygon(s) with the specified offset applied.

    """
    if not isinstance(elements, list):
        elements = [elements]
    polygons_to_offset = []
    for e in elements:
        if isinstance(e, (Component, ComponentReference)):
            polygons_to_offset += e.get_polygons(by_spec=False)
        elif isinstance(e, (Polygon, gdstk.Polygon)):
            polygons_to_offset.append(e)
    if len(polygons_to_offset) == 0:
        return gf.Component("offset")

    layer = gf.get_layer(layer)
    gds_layer, gds_datatype = _parse_layer(layer)
    p = gdstk.offset(
        polygons_to_offset,
        distance=distance,
        join=join,
        tolerance=tolerance,
        precision=precision,
        use_union=use_union,
        layer=gds_layer,
        datatype=gds_datatype,
    )

    component = gf.Component()
    component.add_polygon(p, layer=layer)
    return component


def test_offset() -> None:
    c = gf.components.ring()
    co = offset(c, distance=0.5)
    assert int(co.area()) == 94


if __name__ == "__main__":
    import gdsfactory as gf

    c = gf.Component()
    layer_slab = (2, 0)
    c1 = gf.components.coupler_ring(
        cladding_layers=[layer_slab], cladding_offsets=[0.5]
    )
    d = 0.8
    # d = 1
    c2 = gf.geometry.offset(c1, distance=+d, layer=layer_slab)

    c3 = gf.geometry.offset(c2, distance=-d, layer=layer_slab)

    c << c1.extract(layers=("WG",))
    c << c3
    c.show()



"""You can remap layers."""
from __future__ import annotations

import gdsfactory as gf
from gdsfactory.component import Component


def test_remap_layers() -> Component:
    c = gf.Component("test_remap_layers_sample_device")
    straight = gf.partial(
        gf.components.straight,
        with_bbox=True,
        cladding_layers=None,
        add_pins=None,
        add_bbox=None,
    )

    wg1 = c << straight(length=11, width=1, layer=(1, 0))
    wg2 = c << straight(length=11, width=2, layer=(2, 0))
    wg3 = c << straight(length=11, width=3, layer=(3, 0))

    wg2.connect(port="o1", destination=wg1.ports["o2"])
    wg3.connect(port="o1", destination=wg2.ports["o2"], overlap=1)

    nlayers = len(c.layers)
    assert len(c.layers) == nlayers
    c.remap_layers({(1, 0): (2, 0)})
    assert len(c.layers) == nlayers - 1
    return c


if __name__ == "__main__":
    c = test_remap_layers()
    c.show(show_ports=True)



from itertools import count
from pathlib import Path
import shutil
from srim import TRIM
import os
from itertools import repeat
import pandas as pd


def fragment(step, total):
    remaining = total
    while remaining > 0:
        if step > remaining:
            return remaining
        else:
            remaining -= step
            yield step


def find_folder(directory):
    for i in count():
        path = Path(directory) / str(i)
        if not path.is_dir():
            return str(path.absolute())


def run_fragmented_calculation(
    srim_executable_directory: Path,
    ion,
    target,
    number_ions,
    save_path: Path,
    trim_settings=None,
    step: int = 1000,
):
    """Runs a TRIM calculations in series, with each batch no more than 1000 ions to avoid crashes.

    Arguments:
        srim_executable_directory: directory where running "wine TRIM" opens the software
        ion: pysrim ion
        target: pysrim target
        number_ions: number of ions to simulate
        path: where to save to save data
        trim_settings: dict of SRIM simulation settings
        step: number of simulations per batch. Default 1000.
    """
    for i, num_ions in enumerate(fragment(step, number_ions)):
        print(
            "total ions completed: {:06d}\tion: {}\tions in step: {:06d}".format(
                i * step, ion.symbol, num_ions
            )
        )
        trim_settings = trim_settings or {"calculation": 2}
        trim = TRIM(target, ion, number_ions=num_ions, **trim_settings)
        print(srim_executable_directory)
        trim.run(srim_executable_directory)
        save_directory = find_folder(save_path)
        os.makedirs(save_directory, exist_ok=True)
        TRIM.copy_output_files(srim_executable_directory, save_directory)


def run_parallel_fragmented_calculation(
    srim_executable_directory: Path,
    ion,
    target,
    number_ions,
    save_path: Path,
    trim_settings=None,
    step: int = 1000,
    cores: int = 4,
):
    """Runs "cores" TRIM series calculations in parallel, with each batch no more than 1000 ions to avoid crashes.

    Arguments:
        cores: number of parallel SRIM instances to run.
    """
    from multiprocessing import Pool

    number_ions_per_core = int(number_ions / cores)

    # First deduplicate the TRIM executable to avoid results conflicts
    for i in range(cores):
        if Path(srim_executable_directory.parent / str(i)).exists():
            shutil.rmtree(srim_executable_directory.parent / str(i))
        shutil.copytree(
            srim_executable_directory, srim_executable_directory.parent / str(i)
        )

    with Pool(cores) as p:
        p.starmap(
            run_fragmented_calculation,
            zip(
                [srim_executable_directory.parent / str(i) for i in range(cores)],
                repeat(ion),
                repeat(target),
                repeat(number_ions_per_core),
                [save_path / str(i) for i in range(cores)],
                repeat(trim_settings),
                repeat(step),
            ),
        )

    # Delete redundant executables
    for i in range(cores):
        shutil.rmtree(srim_executable_directory.parent / str(i))


def read_ranges(save_path):
    """Read ranges from a data directory (or subdirectories).

    TODO: contribute back to PySRIM.

    Returns:
        Dataframe with all ion final positions.
        z-column represents depth in the sample
        x-column represent lateral dimension, in tilt angle direction
        y-column represents lateral dimension, perpendicular to tilt angle direction
    """
    dataframes = []
    for root, _dirs, files in os.walk(save_path):
        for filename in files:
            if filename == "RANGE_3D.txt":
                with open(Path(root) / filename, "r", encoding="latin-1") as f:
                    # Find table
                    dataframes.extend(
                        pd.read_csv(f, sep="\s+", header=None)[[1, 2, 3]]  # noqa: W605
                        for num, line in enumerate(f, 1)
                        if "----" in line
                    )
    return pd.concat(dataframes).rename(columns={1: "z", 2: "x", 3: "y"}).reset_index()


if __name__ == "__main__":
    from pathlib import Path
    from srim import Ion, Layer, Target
    import matplotlib.pyplot as plt

    # Define implant
    energy = 1.0e5
    implant = Ion("P", energy=1.0e5)

    # Define layers of target
    nm = 10  # units of SRIM are Angstroms
    um = 1e4
    soi_thickness = 220 * nm
    BOX_thickness = 100 * nm  # instead of 3 * um, ions barely make it to BOX

    # 220nm pure silicon
    soi = Layer(
        {
            "Si": {
                # (float, int, required): Stoichiometry of element (fraction)
                "stoich": 1.0,
                "E_d": 35.0,  # (float, int, optional): Displacement energy [eV]
                # (float, int, optional): Lattice binding energies [eV]. Used for sputtering calculations.
                "lattice": 0.0,
                # (float, int, optional): Surface binding energies [eV]. Used for sputtering calculations.
                "surface": 3.0,
            },
        },
        density=2.3290,  # density [g/cm^3] of material
        width=soi_thickness,  # width [Angstroms] of layer
    )

    # 3um SiO2
    box = Layer(
        {
            "Si": {
                "stoich": 0.33,
            },
            "O": {
                "stoich": 0.67,
            },
        },
        density=2.65,
        width=BOX_thickness,
    )

    # Define multilayer target
    target = Target([soi, box])

    # Simulation parameters
    overwrite = True

    srim_executable_directory = Path("/home/bilodeaus/.wine/drive_c/SRIM")
    srim_data_directory = Path("./tmp/")

    srim_data_directory.mkdir(exist_ok=True, parents=True)
    srim_executable_directory.mkdir(exist_ok=True, parents=True)

    trim_settings = {
        "calculation": 1,
        "angle_ions": 45,
        "ranges": True,
        "plot_mode": 5,
    }

    if overwrite:
        run_parallel_fragmented_calculation(
            srim_executable_directory=srim_executable_directory,
            ion=implant,
            target=target,
            number_ions=10000,
            save_path=srim_data_directory,
            trim_settings=trim_settings,
            step=1000,
            cores=6,
        )

    df = read_ranges(srim_data_directory)

    ax = df.plot.hist(column=["z"], bins=200, alpha=0.5, xlabel="z (A)", density=True)
    ax = df.plot.hist(column=["z"], bins=100, alpha=0.5, xlabel="z (A)", density=True)
    ax = df.plot.hist(column=["x"], bins=100, alpha=0.5, xlabel="x (A)", density=True)
    ax = df.plot.hist(column=["y"], bins=100, alpha=0.5, xlabel="y (A)", density=True)
    plt.show()



from __future__ import annotations

import pathlib
from typing import Optional, Tuple

import gdsfactory.generic_tech as generic
from gdsfactory.component import Component
from gdsfactory.technology import LayerStack, LayerViews
from gdsfactory.typings import Layer


def to_stl(
    component: Component,
    filepath: str,
    layer_views: LayerViews = generic.LAYER_VIEWS,
    layer_stack: LayerStack = generic.LAYER_STACK,
    exclude_layers: Optional[Tuple[Layer, ...]] = None,
) -> None:
    """Exports a Component into STL.

    Args:
        component: to export.
        filepath: to write STL to.
        layer_views: layer colors from Klayout Layer Properties file.
        layer_stack: contains thickness and zmin for each layer.
        exclude_layers: layers to exclude.

    """
    import shapely
    from trimesh.creation import extrude_polygon

    layer_to_thickness = layer_stack.get_layer_to_thickness()
    layer_to_zmin = layer_stack.get_layer_to_zmin()
    filepath = pathlib.Path(filepath)
    exclude_layers = exclude_layers or []

    for layer, polygons in component.get_polygons(by_spec=True).items():
        if (
            layer not in exclude_layers
            and layer in layer_to_thickness
            and layer in layer_to_zmin
        ):
            height = layer_to_thickness[layer]
            zmin = layer_to_zmin[layer]
            filepath_layer = (
                filepath.parent
                / f"{filepath.stem}_{layer[0]}_{layer[1]}{filepath.suffix}"
            )
            for polygon in polygons:
                p = shapely.geometry.Polygon(polygon)
                mesh = extrude_polygon(p, height=height)
                mesh.apply_translation((0, 0, zmin))
                mesh.visual.face_colors = (
                    *layer_views.get_from_tuple(layer).fill_color.as_rgb_tuple(),
                    0.5,
                )
                mesh.export(filepath_layer)


if __name__ == "__main__":
    import gdsfactory as gf
    import gdsfactory.generic_tech as generic

    c = gf.components.taper_strip_to_ridge()
    to_stl(c, layer_views=generic.LAYER_VIEWS, filepath="a.stl")



from __future__ import annotations

import hashlib
import pathlib
from copy import deepcopy
from functools import partial
from pathlib import Path
from typing import Optional

import numpy as np

import gdsfactory as gf
from gdsfactory.generic_tech import LAYER_STACK
from gdsfactory.name import clean_value
from gdsfactory.pdk import get_modes_path
from gdsfactory.typings import ComponentSpec


def get_kwargs_hash(**kwargs) -> str:
    """Returns kwargs parameters hash."""
    kwargs_list = [f"{key}={clean_value(kwargs[key])}" for key in sorted(kwargs.keys())]
    kwargs_string = "_".join(kwargs_list)
    return hashlib.md5(kwargs_string.encode()).hexdigest()[:8]


def _get_modes_path(
    component: ComponentSpec,
    dirpath: Optional[Path] = None,
    extension: Optional[str] = "npz",
    **kwargs,
) -> Path:
    """Return modes npz filepath hashing simulation settings for \
            a consistent unique name.

    Args:
        component: component.
        xsection_bounds: line where to take a cross-sectional mesh for mode solving,
        dirpath: directory to store sparameters in CSV.
            Defaults to active Pdk.sparameters_path.
        extension: to append to the end of the file.
        kwargs: simulation settings.

    """
    dirpath = dirpath or get_modes_path()
    component = gf.get_component(component)

    dirpath = pathlib.Path(dirpath)
    dirpath.mkdir(exist_ok=True, parents=True)
    return dirpath / f"{component.name}_{get_kwargs_hash(**kwargs)}.{extension}"


def _get_modes_data(**kwargs) -> np.ndarray:
    """Returns modes data in a pandas DataFrame.

    Keyword Args:
        cross_section: cross_section or cross_section factory.
        dirpath: directory path to store sparameters.
        kwargs: simulation settings.

    """
    filepath = _get_modes_path(**kwargs)
    return np.load(filepath)


get_modes_path_femwell = partial(_get_modes_path, tool="femwell")


def test_get_modes_path(test: bool = True) -> None:
    import gdsfactory as gf

    nm = 1e-3
    layer_stack2 = deepcopy(LAYER_STACK)
    layer_stack2.layers["core"].thickness = 230 * nm

    cross_section = gf.cross_section.strip()

    p1 = get_modes_path_femwell(cross_section=cross_section)
    p2 = get_modes_path_femwell(cross_section=cross_section, layer_stack=layer_stack2)
    p3 = get_modes_path_femwell(
        cross_section=cross_section, material_name_to_lumerical=dict(si=3.6)
    )

    if test:
        name1 = "xs_adfc05a6_782ce72c"
        name2 = "xs_adfc05a6_a01df6f8"
        name3 = "xs_adfc05a6_e88b3d6d"

        assert p1.stem == name1, p1.stem
        assert p2.stem == name2, p2.stem
        assert p3.stem == name3, p3.stem
    else:
        print(f"name1 = {p1.stem!r}")
        print(f"name2 = {p2.stem!r}")
        print(f"name3 = {p3.stem!r}")


if __name__ == "__main__":
    # cross_section = gf.cross_section.strip(name="strip")
    # p = get_modes_path_femwell(cross_section)
    # print(p)

    # modes = np.load(p)
    # modesd = dict(modes)
    # print(modesd)

    # test_get_modes_path(test=False)
    test_get_modes_path(test=False)



from __future__ import annotations

import warnings
from typing import List, Optional, Tuple, Union

import numpy as np
from numpy import ndarray

import gdsfactory as gf
from gdsfactory.cell import cell
from gdsfactory.component import Component
from gdsfactory.components.mmi1x2 import mmi1x2
from gdsfactory.cross_section import cross_section as cross_section_function
from gdsfactory.port import Port
from gdsfactory.typings import ComponentSpec, Coordinate, CrossSectionSpec, Layer

DEG2RAD = np.pi / 180


def line(
    p_start: Union[Port, Coordinate],
    p_end: Union[Port, Coordinate],
    width: Optional[float] = None,
) -> Tuple[float, float, float, float]:
    if isinstance(p_start, gf.Port):
        width = p_start.width
        p_start = p_start.center

    if isinstance(p_end, gf.Port):
        p_end = p_end.center

    w = width
    angle = np.arctan2(p_end[1] - p_start[1], p_end[0] - p_start[0])
    a = np.pi / 2
    p0 = move_polar_rad_copy(p_start, angle + a, w / 2)
    p1 = move_polar_rad_copy(p_start, angle - a, w / 2)
    p2 = move_polar_rad_copy(p_end, angle - a, w / 2)
    p3 = move_polar_rad_copy(p_end, angle + a, w / 2)
    return p0, p1, p2, p3


def move_polar_rad_copy(pos: Coordinate, angle: float, length: float) -> ndarray:
    """Returns the points of a position (pos) with angle, by shifted by certain.

    length.

    Args:
        pos: position.
        angle: in radians.
        length: extension length in um.
    """
    c = np.cos(angle)
    s = np.sin(angle)
    return pos + length * np.array([c, s])


@cell
def extend_port(port: Port, length: float, layer: Optional[Layer] = None) -> Component:
    """Returns a straight extension component out of a port.

    Args:
        port: port to extend.
        length: extension length in um.
        layer: for the straight section.
    """
    c = Component()
    layer = layer or port.layer

    # Generate a port extension
    p_start = port.center
    angle = port.orientation
    p_end = move_polar_rad_copy(p_start, angle * DEG2RAD, length)
    w = port.width

    _line = line(p_start, p_end, w)

    c.add_polygon(_line, layer=layer)
    c.add_port(name="original", port=port)

    port_settings = port.settings.copy()
    port_settings.update(center=p_end)
    c.add_port(**port_settings)

    return c


@gf.cell
def extend_ports(
    component: ComponentSpec = mmi1x2,
    port_names: Optional[Tuple[str, ...]] = None,
    length: float = 5.0,
    extension: Optional[ComponentSpec] = None,
    port1: Optional[str] = None,
    port2: Optional[str] = None,
    port_type: str = "optical",
    centered: bool = False,
    cross_section: Optional[CrossSectionSpec] = None,
    extension_port_names: Optional[List[str]] = None,
    **kwargs,
) -> Component:
    """Returns a new component with some ports extended.

    You can define extension Spec
    defaults to port cross_section of each port to extend.

    Args:
        component: component to extend ports.
        port_names: list of ports names to extend, if None it extends all ports.
        length: extension length.
        extension: function to extend ports (defaults to a straight).
        port1: extension input port name.
        port2: extension output port name.
        port_type: type of the ports to extend.
        centered: if True centers rectangle at (0, 0).
        cross_section: extension cross_section, defaults to port cross_section
            if port has no cross_section it creates one using width and layer.
        extension_port_names: extension port names add to the new component.

    Keyword Args:
        layer: port GDS layer.
        prefix: port name prefix.
        orientation: in degrees.
        width: port width.
        layers_excluded: List of layers to exclude.
        port_type: optical, electrical, ....
        clockwise: if True, sort ports clockwise, False: counter-clockwise.
    """
    c = gf.Component()
    component = gf.get_component(component)
    cref = c << component
    c.component = component

    if centered:
        cref.x = 0
        cref.y = 0

    ports_all = cref.get_ports_list()
    port_names_all = [p.name for p in ports_all]

    ports_to_extend = cref.get_ports_list(port_type=port_type, **kwargs)
    ports_to_extend_names = [p.name for p in ports_to_extend]
    ports_to_extend_names = port_names or ports_to_extend_names

    for port_name in ports_to_extend_names:
        if port_name not in port_names_all:
            warnings.warn(f"Port Name {port_name!r} not in {port_names_all}")

    for port in ports_all:
        port_name = port.name
        port = cref.ports[port_name]

        if port_name in ports_to_extend_names:
            if extension:
                extension_component = gf.get_component(extension)
            else:
                cross_section_extension = (
                    cross_section
                    or port.cross_section
                    or cross_section_function(layer=port.layer, width=port.width)
                )

                if cross_section_extension is None:
                    raise ValueError("cross_section=None for extend_ports")

                extension_component = gf.components.straight(
                    length=length,
                    cross_section=cross_section_extension,
                )
            port_labels = list(extension_component.ports.keys())
            port1 = port1 or port_labels[0]
            port2 = port2 or port_labels[-1]

            extension_ref = c << extension_component
            extension_ref.connect(port1, port)
            c.add_port(port_name, port=extension_ref.ports[port2])
            extension_port_names = extension_port_names or []
            [
                c.add_port(name, port=extension_ref.ports[name], prefix=port_name)
                for name in extension_port_names
            ]
        else:
            c.add_port(port_name, port=component.ports[port_name])

    c.copy_child_info(component)
    return c


def test_extend_ports() -> Component:
    import gdsfactory.components as pc

    width = 0.5
    xs_strip = gf.partial(
        gf.cross_section.strip,
        width=width,
        cladding_layers=None,
        add_pins=None,
        add_bbox=None,
    )

    c = pc.straight(cross_section=xs_strip)

    c1 = extend_ports(
        component=c,
        cross_section=xs_strip,
    )
    assert len(c.ports) == len(c1.ports)
    p = len(c1.polygons)
    assert p == 0, p
    assert len(c1.references) == 3, len(c1.references)

    c2 = extend_ports(component=c, cross_section=xs_strip, port_names=("o1",))
    p = len(c2.polygons)
    assert p == 0, p
    assert len(c2.references) == 2, len(c2.references)

    c3 = extend_ports(component=c, cross_section=xs_strip)
    p = len(c3.polygons)
    assert p == 0, p

    c4 = extend_ports(component=c, port_type="electrical")
    p = len(c4.polygons)
    assert p == 0, p
    assert len(c4.references) == 1, len(c4.references)

    return c4


__all__ = ["extend_ports", "extend_port"]


if __name__ == "__main__":
    # c0 = gf.components.taper(width2=10)
    extension = gf.components.straight_heater_meander()
    c0 = gf.components.straight()
    # c1 = extend_ports(c0, orientation=0, extension=extension)

    c1 = extend_ports(
        c0,
        extension=extension,
        orientation=0,
        extension_port_names=["e1", "e2"],
        port1="o1",
        port2="o2",
    )
    c1.pprint_ports()
    c1.show(show_ports=True)

    # c = extend_ports(gf.components.mzi_phase_shifter_top_heater_metal)
    # c = test_extend_ports()

    # width = 0.5
    # xs_strip = gf.partial(
    #     gf.cross_section.strip,
    #     width=width,
    #     cladding_layers=None,
    #     add_pins=None,
    #     add_bbox=None,
    # )

    # import gdsfactory.components as pc
    # c = pc.straight(cross_section=xs_strip)
    # c4 = extend_ports(component=c, port_type='electrical')
    # c4.show()



"""PDK stores layers, cross_sections, cell functions ..."""

from __future__ import annotations
import pathlib
import warnings
from functools import partial
from typing import Any, Callable, Optional, Union, Tuple

import numpy as np
from omegaconf import DictConfig
from pydantic import BaseModel, Field, validator

from gdsfactory.config import PATH, logger
from gdsfactory.containers import containers as containers_default
from gdsfactory.events import Event
from gdsfactory.generic_tech import get_generic_pdk
from gdsfactory.materials import MaterialSpec
from gdsfactory.materials import materials_index as materials_index_default
from gdsfactory.read import cell_from_yaml
from gdsfactory.show import show
from gdsfactory.symbols import floorplan_with_block_letters
from gdsfactory.technology import LayerStack, LayerViews
from gdsfactory.typings import (
    CellSpec,
    Component,
    ComponentFactory,
    ComponentSpec,
    CrossSection,
    CrossSectionFactory,
    CrossSectionSpec,
    Dict,
    Layer,
    LayerSpec,
    PathType,
)

component_settings = ["function", "component", "settings"]
cross_section_settings = ["function", "cross_section", "settings"]
layers_required = ["DEVREC", "PORT", "PORTE"]

constants = {
    "fiber_array_spacing": 127.0,
    "fiber_spacing": 50.0,
    "fiber_input_to_output_spacing": 200.0,
    "metal_spacing": 10.0,
}

nm = 1e-3


class GdsWriteSettings(BaseModel):
    """Settings to use when writing to GDS."""

    unit: float = Field(
        default=1e-6,
        description="The units of coordinates in the database. The default is 1e-6 (1 micron).",
    )
    precision: float = Field(
        default=1e-9,
        description="The maximum precision of points in the database. For example, a value of 1e-9 would mean that you have a 1nm grid",
    )
    on_duplicate_cell: str = Field(
        default="warn",
        description="""Action to take when a duplicate cell is encountered on gds write (usually problematic). Options are
                        "warn" (default): overwrite all duplicate cells with one of the duplicates (arbitrarily).
                        "error": throw a ValueError when attempting to write a gds with duplicate cells.
                        "overwrite": overwrite all duplicate cells with one of the duplicates, without warning.""",
    )
    flatten_invalid_refs: bool = Field(
        default=False,
        description="If true, will auto-correct (and flatten) cell references which are off-grid or rotated by non-manhattan angles.",
    )
    max_points: int = Field(
        default=4000,
        description="Maximum number of points to allow in a polygon before fracturing.",
    )


class OasisWriteSettings(BaseModel):
    compression_level: int = Field(
        default=6,
        description="Level of compression for cells (between 0 and 9). Setting to 0 will disable cell compression, 1 gives the best speed, and 9 gives the best compression.",
    )
    detect_rectangles: bool = Field(
        default=True, description="If true, stores rectangles in a compressed format."
    )
    detect_trapezoids: bool = Field(
        default=True, description="If true, stores trapezoids in a compressed format."
    )
    circle_tolerance: bool = Field(
        default=0,
        description="Tolerance for detecting circles. If less or equal to 0, no detection is performed. Circles are stored in compressed format.",
    )
    validation: Optional[str] = Field(
        default=None,
        description="Type of validation to include in the saved file ('crc32', 'checksum32', or None).",
    )
    standard_properties: bool = Field(
        default=False,
        description="If true, stores standard OASIS properties in the file.",
    )


class Pdk(BaseModel):
    """Store layers, cross_sections, cell functions, simulation_settings ...

    only one Pdk can be active at a given time.

    Parameters:
        name: PDK name.
        cross_sections: dict of cross_sections factories.
        cells: dict of parametric cells that return Components.
        symbols: dict of symbols names to functions.
        default_symbol_factory:
        containers: dict of pcells that contain other cells.
        base_pdk: a pdk to copy from and extend.
        default_decorator: decorate all cells, if not otherwise defined on the cell.
        layers: maps name to gdslayer/datatype.
            For example dict(si=(1, 0), sin=(34, 0)).
        layer_stack: maps name to layer numbers, thickness, zmin, sidewall_angle.
            if can also contain material properties
            (refractive index, nonlinear coefficient, sheet resistance ...).
        layer_views: includes layer name to color, opacity and pattern.
        sparameters_path: to store Sparameters simulations.
        modes_path: to store Sparameters simulations.
        interconnect_cml_path: path to interconnect CML (optional).
        grid_size: in um. Defaults to 1nm.
        warn_off_grid_ports: raises warning when extruding paths with offgrid ports.
            For example, if you try to create a waveguide with 1.5nm length.
        constants: dict of constants for the PDK.
        bend_points_distance: default points distance for bends in um.

    """

    name: str
    cross_sections: Dict[str, CrossSectionFactory] = Field(default_factory=dict)
    cells: Dict[str, ComponentFactory] = Field(default_factory=dict)
    symbols: Dict[str, ComponentFactory] = Field(default_factory=dict)
    default_symbol_factory: Callable = floorplan_with_block_letters
    containers: Dict[str, ComponentFactory] = containers_default
    base_pdk: Optional[Pdk] = None
    default_decorator: Optional[Callable[[Component], None]] = None
    layers: Dict[str, Layer] = Field(default_factory=dict)
    layer_stack: Optional[LayerStack] = None
    layer_views: Optional[LayerViews] = None
    layer_transitions: Dict[Union[Layer, Tuple[Layer, Layer]], ComponentSpec] = Field(
        default_factory=dict
    )
    sparameters_path: Optional[PathType] = None
    modes_path: Optional[PathType] = PATH.modes
    interconnect_cml_path: Optional[PathType] = None
    warn_off_grid_ports: bool = False
    constants: Dict[str, Any] = constants
    materials_index: Dict[str, MaterialSpec] = materials_index_default
    routing_strategies: Optional[Dict[str, Callable]] = None
    circuit_yaml_parser: Callable = cell_from_yaml
    gds_write_settings: GdsWriteSettings = GdsWriteSettings()
    oasis_settings: OasisWriteSettings = OasisWriteSettings()
    bend_points_distance: float = 20 * nm

    @property
    def grid_size(self):
        """The minimum unit resolvable on the layout grid, relative to the unit."""
        return self.gds_write_settings.precision / self.gds_write_settings.unit

    @grid_size.setter
    def grid_size(self, value):
        self.gds_write_settings.precision = value * self.gds_write_settings.unit

    class Config:
        """Configuration."""

        extra = "forbid"
        fields = {
            "cross_sections": {"exclude": True},
            "cells": {"exclude": True},
            "containers": {"exclude": True},
            "default_decorator": {"exclude": True},
        }

    @validator("sparameters_path")
    def is_pathlib_path(cls, path):
        return pathlib.Path(path)

    def validate_layers(self):
        for layer in layers_required:
            if layer not in self.layers:
                raise ValueError(
                    f"{layer!r} not in Pdk.layers {list(self.layers.keys())}"
                )

    def activate(self) -> None:
        """Set current pdk to as the active pdk."""
        from gdsfactory.cell import clear_cache

        logger.info(f"{self.name!r} PDK is now active")

        clear_cache()

        if self.base_pdk:
            cross_sections = self.base_pdk.cross_sections
            cross_sections.update(self.cross_sections)
            self.cross_sections = cross_sections

            cells = self.base_pdk.cells
            cells.update(self.cells)
            self.cells.update(cells)

            containers = self.base_pdk.containers
            containers.update(self.containers)
            self.containers.update(containers)

            layers = self.base_pdk.layers
            layers.update(self.layers)
            self.layers.update(layers)

            if not self.default_decorator:
                self.default_decorator = self.base_pdk.default_decorator
        self.validate_layers()
        _set_active_pdk(self)

    def register_cells(self, **kwargs) -> None:
        """Register cell factories."""
        for name, cell in kwargs.items():
            if not callable(cell):
                raise ValueError(
                    f"{cell} is not callable, make sure you register "
                    "cells functions that return a Component"
                )
            if name in self.cells:
                warnings.warn(f"Overwriting cell {name!r}")

            self.cells[name] = cell
            on_cell_registered.fire(name=name, cell=cell, pdk=self)

    def register_containers(self, **kwargs) -> None:
        """Register container factories."""
        for name, cell in kwargs.items():
            if not callable(cell):
                raise ValueError(
                    f"{cell} is not callable, make sure you register "
                    "cells functions that return a Component"
                )
            if name in self.containers:
                warnings.warn(f"Overwriting container {name!r}")

            self.containers[name] = cell
            on_container_registered.fire(name=name, cell=cell, pdk=self)

    def register_cross_sections(self, **kwargs) -> None:
        """Register cross_sections factories."""
        for name, cross_section in kwargs.items():
            if not callable(cross_section):
                raise ValueError(
                    f"{cross_section} is not callable, make sure you register "
                    "cross_section functions that return a CrossSection"
                )
            if name in self.cross_sections:
                warnings.warn(f"Overwriting cross_section {name!r}")
            self.cross_sections[name] = cross_section
            on_cross_section_registered.fire(
                name=name, cross_section=cross_section, pdk=self
            )

    def register_cells_yaml(
        self,
        dirpath: Optional[PathType] = None,
        update: bool = False,
        **kwargs,
    ) -> None:
        """Load *.pic.yml YAML files and register them as cells.

        Args:
            dirpath: directory to recursive search for YAML cells.
            update: does not raise ValueError if cell already registered.

        Keyword Args:
            cell_name: cell function. To update cells dict.

        """

        message = "Updated" if update else "Registered"

        if dirpath:
            dirpath = pathlib.Path(dirpath)

            if not dirpath.is_dir():
                raise ValueError(f"{dirpath!r} needs to be a directory.")

            for filepath in dirpath.glob("*/**/*.pic.yml"):
                name = filepath.stem.split(".")[0]
                if not update and name in self.cells:
                    raise ValueError(
                        f"ERROR: Cell name {name!r} from {filepath} already registered."
                    )
                parser = self.circuit_yaml_parser
                self.cells[name] = parser(filepath, name=name)
                on_yaml_cell_registered.fire(name=name, cell=self.cells[name], pdk=self)
                logger.info(f"{message} cell {name!r}")

        for k, v in kwargs.items():
            if not update and k in self.cells:
                raise ValueError(f"ERROR: Cell name {k!r} already registered.")
            self.cells[k] = v
            logger.info(f"{message} cell {k!r}")

    def remove_cell(self, name: str):
        """Removes cell from a PDK."""
        if name not in self.cells:
            raise ValueError(f"{name!r} not in {list(self.cells.keys())}")
        self.cells.pop(name)
        logger.info(f"Removed cell {name!r}")

    def get_cell(self, cell: CellSpec, **kwargs) -> ComponentFactory:
        """Returns ComponentFactory from a cell spec."""
        cells_and_containers = set(self.cells.keys()).union(set(self.containers.keys()))

        if callable(cell):
            return cell
        elif isinstance(cell, str):
            if cell not in cells_and_containers:
                cells = list(self.cells.keys())
                containers = "\n".join(list(self.containers.keys()))
                raise ValueError(
                    f"{cell!r} from PDK {self.name!r} not in cells: {cells} "
                    f"or containers: {containers}"
                )
            cell = self.cells[cell] if cell in self.cells else self.containers[cell]
            return cell
        elif isinstance(cell, (dict, DictConfig)):
            for key in cell.keys():
                if key not in component_settings:
                    raise ValueError(
                        f"Invalid setting {key!r} not in {component_settings}"
                    )
            settings = dict(cell.get("settings", {}))
            settings.update(**kwargs)

            cell_name = cell.get("function")
            if not isinstance(cell_name, str) or cell_name not in cells_and_containers:
                cells = list(self.cells.keys())
                containers = list(self.containers.keys())
                raise ValueError(
                    f"{cell_name!r} from PDK {self.name!r} not in cells: {cells} "
                    f"or containers: {containers}"
                )
            cell = (
                self.cells[cell_name]
                if cell_name in self.cells
                else self.containers[cell_name]
            )
            return partial(cell, **settings)
        else:
            raise ValueError(
                "get_cell expects a CellSpec (ComponentFactory, string or dict),"
                f"got {type(cell)}"
            )

    def get_component(self, component: ComponentSpec, **kwargs) -> Component:
        """Returns component from a component spec."""
        return self._get_component(
            component=component, cells=self.cells, containers=self.containers, **kwargs
        )

    def get_symbol(self, component: ComponentSpec, **kwargs) -> Component:
        """Returns a component's symbol from a component spec."""
        # this is a pretty rough first implementation
        try:
            self._get_component(
                component=component, cells=self.symbols, containers={}, **kwargs
            )
        except ValueError:
            component = self.get_component(component, **kwargs)
            return self.default_symbol_factory(component)

    def _get_component(
        self,
        component: ComponentSpec,
        cells: Dict[str, Callable],
        containers: Dict[str, Callable],
        **kwargs,
    ) -> Component:
        """Returns component from a component spec."""
        cells_and_containers = set(cells.keys()).union(set(containers.keys()))

        if isinstance(component, Component):
            if kwargs:
                raise ValueError(f"Cannot apply kwargs {kwargs} to {component.name!r}")
            return component
        elif callable(component):
            return component(**kwargs)
        elif isinstance(component, str):
            if component not in cells_and_containers:
                cells = list(cells.keys())
                containers = list(containers.keys())
                raise ValueError(
                    f"{component!r} not in PDK {self.name!r} cells: {cells} "
                    f"or containers: {containers}"
                )
            cell = cells[component] if component in cells else containers[component]
            return cell(**kwargs)
        elif isinstance(component, (dict, DictConfig)):
            for key in component.keys():
                if key not in component_settings:
                    raise ValueError(
                        f"Invalid setting {key!r} not in {component_settings}"
                    )
            settings = dict(component.get("settings", {}))
            settings.update(**kwargs)

            cell_name = component.get("component", None)
            cell_name = cell_name or component.get("function")
            if not isinstance(cell_name, str) or cell_name not in cells_and_containers:
                cells = list(cells.keys())
                containers = list(containers.keys())
                raise ValueError(
                    f"{cell_name!r} from PDK {self.name!r} not in cells: {cells} "
                    f"or containers: {containers}"
                )
            cell = cells[cell_name] if cell_name in cells else containers[cell_name]
            component = cell(**settings)
            return component
        else:
            raise ValueError(
                "get_component expects a ComponentSpec (Component, ComponentFactory, "
                f"string or dict), got {type(component)}"
            )

    def get_cross_section(
        self, cross_section: CrossSectionSpec, **kwargs
    ) -> CrossSection:
        """Returns cross_section from a cross_section spec."""
        if isinstance(cross_section, CrossSection):
            return cross_section.copy(**kwargs)
        elif callable(cross_section):
            return cross_section(**kwargs)
        elif isinstance(cross_section, str):
            if cross_section not in self.cross_sections:
                cross_sections = list(self.cross_sections.keys())
                raise ValueError(f"{cross_section!r} not in {cross_sections}")
            cross_section_factory = self.cross_sections[cross_section]
            return cross_section_factory(**kwargs)
        elif isinstance(cross_section, (dict, DictConfig)):
            for key in cross_section.keys():
                if key not in cross_section_settings:
                    raise ValueError(
                        f"Invalid setting {key!r} not in {cross_section_settings}"
                    )
            cross_section_factory_name = cross_section.get("cross_section", None)
            cross_section_factory_name = (
                cross_section_factory_name or cross_section.get("function")
            )
            if (
                not isinstance(cross_section_factory_name, str)
                or cross_section_factory_name not in self.cross_sections
            ):
                cross_sections = list(self.cross_sections.keys())
                raise ValueError(
                    f"{cross_section_factory_name!r} not in {cross_sections}"
                )
            cross_section_factory = self.cross_sections[cross_section_factory_name]
            settings = dict(cross_section.get("settings", {}))
            settings.update(**kwargs)

            return cross_section_factory(**settings)
        else:
            raise ValueError(
                "get_cross_section expects a CrossSectionSpec (CrossSection, "
                f"CrossSectionFactory, string or dict), got {type(cross_section)}"
            )

    def get_layer(self, layer: LayerSpec) -> Layer:
        """Returns layer from a layer spec."""
        if isinstance(layer, (tuple, list)):
            if len(layer) != 2:
                raise ValueError(f"{layer!r} needs two integer numbers.")
            return layer
        elif isinstance(layer, int):
            return (layer, 0)
        elif isinstance(layer, str):
            if layer not in self.layers:
                raise ValueError(f"{layer!r} not in {self.layers.keys()}")
            return self.layers[layer]
        elif layer is np.nan:
            return np.nan
        elif layer is None:
            return
        else:
            raise ValueError(
                f"{layer!r} needs to be a LayerSpec (string, int or Layer)"
            )

    def get_layer_views(self) -> LayerViews:
        if self.layer_views is None:
            raise ValueError(f"layer_views for Pdk {self.name!r} is None")
        return self.layer_views

    def get_layer_stack(self) -> LayerStack:
        if self.layer_stack is None:
            raise ValueError(f"layer_stack for Pdk {self.name!r} is None")
        return self.layer_stack

    def get_constant(self, key: str) -> Any:
        if not isinstance(key, str):
            return key
        if key not in self.constants:
            constants = list(self.constants.keys())
            raise ValueError(f"{key!r} not in {constants}")
        return self.constants[key]

    def get_material_index(self, key: str, *args, **kwargs) -> float:
        if key not in self.materials_index:
            material_names = list(self.materials_index.keys())
            raise ValueError(f"{key!r} not in {material_names}")
        material = self.materials_index[key]
        return material(*args, **kwargs) if callable(material) else material

    # _on_cell_registered = Event()
    # _on_container_registered: Event = Event()
    # _on_yaml_cell_registered: Event = Event()
    # _on_cross_section_registered: Event = Event()
    #
    # @property
    # def on_cell_registered(self) -> Event:
    #     return self._on_cell_registered
    #
    # @property
    # def on_container_registered(self) -> Event:
    #     return self._on_container_registered
    #
    # @property
    # def on_yaml_cell_registered(self) -> Event:
    #     return self._on_yaml_cell_registered
    #
    # @property
    # def on_cross_section_registered(self) -> Event:
    #     return self._on_cross_section_registered


GENERIC_PDK = get_generic_pdk()
_ACTIVE_PDK = GENERIC_PDK


def get_material_index(material: MaterialSpec, *args, **kwargs) -> Component:
    return _ACTIVE_PDK.get_material_index(material, *args, **kwargs)


def get_component(component: ComponentSpec, **kwargs) -> Component:
    return _ACTIVE_PDK.get_component(component, **kwargs)


def get_cell(cell: CellSpec, **kwargs) -> ComponentFactory:
    return _ACTIVE_PDK.get_cell(cell, **kwargs)


def get_cross_section(cross_section: CrossSectionSpec, **kwargs) -> CrossSection:
    return _ACTIVE_PDK.get_cross_section(cross_section, **kwargs)


def get_layer(layer: LayerSpec) -> Layer:
    return _ACTIVE_PDK.get_layer(layer)


def get_layer_views() -> LayerViews:
    return _ACTIVE_PDK.get_layer_views()


def get_layer_stack() -> LayerStack:
    return _ACTIVE_PDK.get_layer_stack()


def get_active_pdk() -> Pdk:
    return _ACTIVE_PDK


def get_grid_size() -> float:
    return _ACTIVE_PDK.grid_size


def get_constant(constant_name: Any) -> Any:
    """If constant_name is a string returns a the value from the dict."""
    return _ACTIVE_PDK.get_constant(constant_name)


def get_sparameters_path() -> pathlib.Path:
    if _ACTIVE_PDK.sparameters_path is None:
        raise ValueError(f"{_ACTIVE_PDK.name!r} has no sparameters_path")
    return _ACTIVE_PDK.sparameters_path


def get_modes_path() -> Optional[pathlib.Path]:
    return _ACTIVE_PDK.modes_path


def get_interconnect_cml_path() -> pathlib.Path:
    if _ACTIVE_PDK.interconnect_cml_path is None:
        raise ValueError(f"{_ACTIVE_PDK.name!r} has no interconnect_cml_path")
    return _ACTIVE_PDK.interconnect_cml_path


def _set_active_pdk(pdk: Pdk) -> None:
    global _ACTIVE_PDK
    old_pdk = _ACTIVE_PDK
    _ACTIVE_PDK = pdk
    on_pdk_activated.fire(old_pdk=old_pdk, new_pdk=pdk)


def get_routing_strategies() -> Dict[str, Callable]:
    """Gets a dictionary of named routing functions available to the PDK, if defined, or gdsfactory defaults otherwise."""
    from gdsfactory.routing.factories import (
        routing_strategy as default_routing_strategies,
    )

    routing_strategies = get_active_pdk().routing_strategies
    if routing_strategies is None:
        routing_strategies = default_routing_strategies
    return routing_strategies


on_pdk_activated: Event = Event()
on_cell_registered: Event = Event()
on_container_registered: Event = Event()
on_yaml_cell_registered: Event = Event()
on_yaml_cell_modified: Event = Event()
on_cross_section_registered: Event = Event()

on_container_registered.add_handler(on_cell_registered.fire)
on_yaml_cell_registered.add_handler(on_cell_registered.fire)
on_yaml_cell_modified.add_handler(show)


if __name__ == "__main__":
    from gdsfactory.components import cells
    from gdsfactory.cross_section import cross_sections

    # c = _ACTIVE_PDK.get_component("straight")
    # print(c.settings)
    # on_pdk_activated += print
    # set_active_pdk(GENERIC)
    c = Pdk(
        name="demo",
        cells=cells,
        cross_sections=cross_sections,
        # layers=dict(DEVREC=(3, 0), PORTE=(3, 5)),
        sparameters_path="/home",
    )
    print(c.layers)



from __future__ import annotations

from enum import Enum
from typing import Any, Dict, List, Optional, Union

import yaml
from pydantic import AnyUrl, BaseModel, Extra, Field

import gdsfactory as gf


class CrossSection(BaseModel):
    __root__: str = Field(
        ..., description="A cross section to use for waveguides or traces."
    )


class RouteSettings(BaseModel):
    cross_section: Optional[CrossSection] = None
    separation: Optional[float] = Field(
        5.0, description="The minimum separation between routes in the bundle [um]."
    )

    class Config:
        """Extra config."""

        extra = Extra.allow


class RoutingStrategy(BaseModel):
    __root__: str = Field(..., description="The type of routing to use")


class Links(BaseModel):
    pass

    class Config:
        """Extra config."""

        extra = Extra.allow


class PortEnum(Enum):
    ne = "ne"
    nc = "nc"
    nw = "nw"
    se = "se"
    sc = "sc"
    sw = "sw"
    ce = "ce"
    cw = "cw"
    cc = "cc"
    center = "center"


class Placement(BaseModel):
    class Config:
        """Extra config."""

        extra = Extra.forbid

    x: Optional[Union[str, float]] = Field(
        None,
        description="The x location at which to place the component.\nThis can either be a number or an other_inst,port definition, meaning it will be placed relative to the port specified on the other instance. \nIf port keyword is defined, this will be relative to the specified port. Otherwise, it will be relative to the cell origin.",
    )
    y: Optional[Union[str, float]] = Field(
        None,
        description="The y location at which to place the component.\nThis can either be a number or an other_inst,port definition, meaning it will be placed relative to the port specified on the other instance. \nIf port keyword is defined, this will be relative to the specified port. Otherwise, it will be relative to the cell origin.",
    )
    port: Optional[Union[str, PortEnum]] = Field(
        None,
        description="The port or keyword used to anchor the component. Either specify any port on the instance or one of these special keywords:\nne, nc, nw, se, sc, sw, ce, cw, cc for the northeast, north-center, northwest, etc. coordinates of the cell",
    )
    rotation: Optional[float] = Field(
        0,
        description="The rotation of the cell about the origin, or port if defined.",
    )
    dx: Optional[float] = Field(
        None,
        description="An additional displacement in the x direction. Useful if x is defined using other_inst,port syntax",
    )
    dy: Optional[float] = Field(
        None,
        description="An additional displacement in the y direction. Useful if y is defined using other_inst,port syntax",
    )
    mirror: Optional[bool] = Field(
        None,
        description="true/false value indicating whether we should flip horizontally.",
    )


class Instance(BaseModel):
    class Config:
        """Extra config."""

        extra = Extra.forbid

    component: str
    settings: Optional[Dict[str, Any]] = Field(
        None, description="Settings for the component"
    )


class Route(BaseModel):
    class Config:
        """Extra config."""

        extra = Extra.forbid

    routing_strategy: Optional[RoutingStrategy] = None
    settings: Optional[RouteSettings] = None
    links: Dict[str, str]


class PicYamlConfiguration(BaseModel):
    _schema: Optional[AnyUrl] = Field(None, alias="$schema")
    instances: Optional[Dict[str, Instance]] = None
    placements: Optional[Dict[str, Placement]] = None
    routes: Optional[Dict[str, Route]] = None
    ports: Optional[Dict[str, str]] = None

    def add_instance(
        self, name: str, component: gf.Component, placement: Optional[Placement] = None
    ):
        component_name = component.settings.function_name
        component_settings = component.settings.changed
        self.instances[name] = Instance(
            component=component_name, settings=component_settings
        )
        if not placement:
            placement = Placement()
        self.placements[name] = placement

    def move_instance_to(self, name: str, x: float, y: float):
        self.placements[name].x = x
        self.placements[name].y = y

    def move_instance(self, name: str, dx: float, dy: float):
        if not self.placements[name].dx:
            self.placements[name].dx = 0
        self.placements[name].dx += dx
        if not self.placements[name].dy:
            self.placements[name].dy = 0
        self.placements[name].dy += dy

    def to_yaml(self, filename):
        netlist = self.dict()
        with open(filename, mode="w") as f:
            yaml.dump(netlist, f, default_flow_style=None, sort_keys=False)


class SchematicConfiguration(BaseModel):
    _schema: Optional[AnyUrl] = Field(None, alias="$schema")
    instances: Optional[Dict[str, Instance]] = None
    schematic_placements: Optional[Dict[str, Placement]] = None
    nets: Optional[List[List[str]]] = None
    ports: Optional[Dict[str, str]] = None

    @property
    def placements(self):
        return self.schematic_placements

    def add_instance(
        self,
        name: str,
        component: Union[str, gf.Component],
        placement: Optional[Placement] = None,
        settings: Dict[str, Any] = None,
    ):
        if isinstance(component, gf.Component):
            component_name = component.settings.function_name
            component_settings = component.settings.changed
            if settings:
                component_settings = component_settings | settings
            self.instances[name] = Instance(
                component=component_name, settings=component_settings
            )
        else:
            if not settings:
                settings = {}
            self.instances[name] = Instance(component=component, settings=settings)
        if name not in self.placements:
            if not placement:
                placement = Placement()
            self.placements[name] = placement


def create_pic():
    return PicYamlConfiguration(
        instances={},
        placements={},
        routes={},
        ports={},
    )



try:
    from ipywidgets import HTML, Image
    from ipyevents import Event
    from typing import Optional
    import klayout.db as db
    import klayout.lay as lay
except ImportError as e:
    print(
        "You need install jupyter notebook plugin with `pip install gdsfactory[full]`"
    )
    raise e


class LayoutViewer:
    def __init__(self, filepath: str, layer_properties: Optional[str]):
        filepath = str(filepath)
        layer_properties = str(layer_properties)
        self.filepath = filepath
        self.layer_properties = layer_properties
        self.layout_view = lay.LayoutView()
        self.load_layout(filepath, layer_properties)
        pixel_buffer = self.layout_view.get_pixels_with_options(800, 600)
        png_data = pixel_buffer.to_png_data()
        self.image = Image(value=png_data, format="png")
        scroll_event = Event(source=self.image, watched_events=["wheel"])
        scroll_event.on_dom_event(self.on_scroll)
        self.wheel_info = HTML("Waiting for a scroll...")
        self.mouse_info = HTML("Waiting for a mouse event...")
        # self.layout_view.on_image_updated_event = lambda: self.refresh
        mouse_event = Event(
            source=self.image, watched_events=["mousedown", "mouseup", "mousemove"]
        )
        mouse_event.on_dom_event(self.on_mouse_down)

    def load_layout(self, filepath: str, layer_properties: Optional[str]):
        """Loads a GDS layout.

        Args:
            filepath: path for the GDS layout.
            layer_properties: Optional path for the layer_properties klayout file (lyp).
        """
        self.layout_view.load_layout(filepath)
        self.layout_view.max_hier()
        if layer_properties:
            self.layout_view.load_layer_props(layer_properties)

    def refresh(self):
        pixel_buffer = self.layout_view.get_pixels_with_options(800, 600)
        png_data = pixel_buffer.to_png_data()
        self.image.value = png_data

    def _get_modifier_buttons(self, event):
        shift = event["shiftKey"]
        alt = event["altKey"]
        ctrl = event["ctrlKey"]
        # meta = event["metaKey"]

        mouse_buttons = event["buttons"]

        buttons = 0
        if shift:
            buttons |= lay.ButtonState.ShiftKey
        if alt:
            buttons |= lay.ButtonState.AltKey
        if ctrl:
            buttons |= lay.ButtonState.ControlKey

        if mouse_buttons == 1:
            buttons |= lay.ButtonState.LeftButton
        elif mouse_buttons == 2:
            buttons |= lay.ButtonState.RightButton
        elif mouse_buttons == 4:
            buttons |= lay.ButtonState.MidButton

        return buttons

    def on_scroll(self, event):
        delta = event["deltaY"]
        # x = event["offsetX"]
        # y = event["offsetY"]
        self.wheel_info.value = f"scroll event: {event}"
        # buttons = self._get_modifier_buttons(event)
        # TODO: this is what I *want* to respond with, but it doesn't work, so I am using zoom_in/zoom_out instead
        # self.layout_view.send_wheel_event(delta, False, db.Point(x, y), buttons)
        if delta < 0:
            self.layout_view.zoom_in()
        else:
            self.layout_view.zoom_out()
        self.refresh()

    def on_mouse_down(self, event):
        x = event["offsetX"]
        y = event["offsetY"]
        moved_x = event["movementX"]
        moved_y = event["movementY"]
        buttons = self._get_modifier_buttons(event)
        # TODO: this part is also not working. why?
        if event == "mousedown":
            self.layout_view.send_mouse_press_event(db.Point(x, y), buttons)
        elif event == "mouseup":
            self.layout_view.send_mouse_release_event(db.Point(x, y), buttons)
        elif event == "mousemove":
            self.layout_view.send_mouse_move_event(db.Point(moved_x, moved_y), buttons)
        self.refresh()
        self.mouse_info.value = f"mouse event: {event}"



from __future__ import annotations

from simphony.libraries.siepic import Waveguide

from gdsfactory.config import logger


def straight(
    length: float = 10.0,
    width: float = 0.5,
    thickness: float = 0.22,
    sw_angle: float = 90.0,
    **kwargs,
):
    """Return simphony Model for a Straight straight.

    Args:
        length: Length of the straight in um.
        width: Width of the straight in um (Valid for 0.4-0.6).
        thickness: Thickness of straight in um (Valid for 180nm-240nm).
        sw_angle: Sidewall angle. Valid for 80-90 degrees.
        kwargs: geometrical args that this model ignores

    """
    logger.info(f"ignoring {kwargs.keys()}")
    width *= 1e-6
    thickness *= 1e-6
    length *= 1e-6

    model = Waveguide(
        width=width, thickness=thickness, sw_angle=sw_angle, length=length
    )
    model.rename_pins("o1", "o2")
    return model


if __name__ == "__main__":
    import matplotlib.pyplot as plt

    from gdsfactory.simulation.simphony.plot_model import plot_model

    c = straight()

    plot_model(c, logscale=False)
    plt.show()



"""Generate the code from a GDS file based PDK."""

from __future__ import annotations

import datetime
import pathlib
from pathlib import Path
from typing import Dict, Optional

import gdstk

from gdsfactory.component import _timestamp2019
from gdsfactory.config import PATH, logger
from gdsfactory.name import clean_name
from gdsfactory.read.import_gds import import_gds
from gdsfactory.typings import PathType

script_prefix = """
from pathlib import PosixPath
from functools import partial
import gdsfactory as gf

add_ports_optical = gf.partial(
    gf.add_ports.add_ports_from_markers_inside, pin_layer=(1, 0), port_layer=(2, 0)
)
add_ports_electrical = gf.partial(
    gf.add_ports.add_ports_from_markers_inside, pin_layer=(41, 0), port_layer=(1, 0)
)
add_ports = gf.compose(add_ports_optical, add_ports_electrical)

"""


def get_script(gdspath: PathType, module: Optional[str] = None) -> str:
    """Returns script for importing a fixed cell.

    Args:
        gdspath: fixed cell gdspath.
        module: if any includes plot directive.

    """
    cell = clean_name(gdspath.stem)
    gdspath = gdspath.stem + gdspath.suffix

    package = module.split(".")[0] if module and "." in module else module
    if module:
        return f"""

@gf.cell
def {cell}()->gf.Component:
    '''Returns {cell} fixed cell.

    .. plot::
      :include-source:

      import {package}

      c = {module}.{cell}()
      c.plot()
    '''
    return import_gds({str(gdspath)!r})

"""

    else:
        return f"""

@gf.cell
def {cell}()->gf.Component:
    '''Returns {cell} fixed cell.'''
    return import_gds({str(gdspath)!r})

"""


def get_import_gds_script(dirpath: PathType, module: Optional[str] = None) -> str:
    """Returns import_gds script from a directory with all the GDS files.

    Args:
        dirpath: fixed cell directory path.
        module: Optional plot directive to plot imported component.

    """
    dirpath = pathlib.Path(dirpath)
    if not dirpath.exists():
        raise ValueError(f"{dirpath.absolute()!r} does not exist.")

    gdspaths = list(dirpath.glob("*.gds")) + list(dirpath.glob("*.GDS"))

    if not gdspaths:
        raise ValueError(f"No GDS files found at {dirpath.absolute()!r}.")

    logger.info(f"Writing {len(gdspaths)} cells from {dirpath.absolute()!r}")

    script = [script_prefix]
    script += [f"gdsdir = {dirpath.absolute()!r}\n"]
    script += [
        "import_gds = partial(gf.import_gds, gdsdir=gdsdir, decorator=add_ports)\n"
    ]

    cells = [get_script(gdspath, module=module) for gdspath in gdspaths]
    script += sorted(cells)
    return "\n".join(script)


def write_cells_recursively(
    cell: gdstk.Cell,
    unit: float = 1e-6,
    precision: float = 1e-9,
    timestamp: Optional[datetime.datetime] = _timestamp2019,
    dirpath: Optional[pathlib.Path] = None,
) -> Dict[str, Path]:
    """Write gdstk cells recursively.

    Args:
        cell: gdstk cell.
        unit: unit size for objects in library. 1um by default.
        precision: for library dimensions (m). 1nm by default.
        timestamp: Defaults to 2019-10-25. If None uses current time.
        dirpath: directory for the GDS file to write to.

    Returns:
        gdspaths: dict of cell name to gdspath.

    """
    dirpath = dirpath or pathlib.Path.cwd()
    gdspaths = {}

    for c in cell.dependencies(True):
        gdspath = dirpath / f"{c.name}.gds"

        lib = gdstk.Library(unit=unit, precision=precision)
        lib.add(cell)
        lib.add(*cell.dependencies(True))
        lib.write_gds(gdspath)
        logger.info(f"Write {cell.name!r} to {gdspath}")

        gdspaths[c.name] = gdspath

    return gdspaths


def write_cells(
    gdspath: Optional[PathType] = None,
    dirpath: Optional[PathType] = None,
    unit: float = 1e-6,
    precision: float = 1e-9,
    timestamp: Optional[datetime.datetime] = _timestamp2019,
    recursively: bool = False,
    flatten: bool = False,
) -> Dict[str, Path]:
    """Writes cells into separate GDS files.

    Args:
        gdspath: GDS file to write cells.
        dirpath: directory path to write GDS files to.
            Defaults to current working directory.
        unit: unit size for objects in library. 1um by default.
        precision: for object dimensions in the library (m). 1nm by default.
        timestamp: Defaults to 2019-10-25. If None uses current time.
        recursively: writes all cells recursively. If False writes only top cells.
        flatten: flatten cell.

    Returns:
        gdspaths: dict of cell name to gdspath.

    """
    cell = gdstk.read_gds(gdspath)
    top_level_cells = cell.top_level()
    top_cellnames = [c.name for c in top_level_cells]

    dirpath = dirpath or pathlib.Path.cwd()
    dirpath = pathlib.Path(dirpath).absolute()
    dirpath.mkdir(exist_ok=True, parents=True)

    gdspaths = {}
    components = {}

    for cellname in top_cellnames:
        c = import_gds(gdspath=gdspath, cellname=cellname)
        if flatten:
            c = c.flatten()
        components[cellname] = c

    for component_name, component in components.items():
        gdspath = dirpath / f"{component_name}.gds"
        component.write_gds(gdspath)
        gdspaths[component_name] = gdspath

    if recursively:
        for cell in top_level_cells:
            if flatten:
                cell = cell.flatten()

            gdspath = dirpath / f"{cell.name}.gds"

            lib = gdstk.Library(unit=unit, precision=precision)
            lib.add(cell)
            lib.add(*cell.dependencies(True))
            lib.write_gds(gdspath)

            logger.info(f"Write {cell.name!r} to {gdspath}")
            gdspaths[cell.name] = gdspath

            gdspaths2 = write_cells_recursively(
                cell=cell,
                unit=unit,
                precision=precision,
                timestamp=timestamp,
                dirpath=dirpath,
            )
            gdspaths.update(gdspaths2)
    return gdspaths


def test_write_cells_recursively():
    gdspath = PATH.gdsdir / "mzi2x2.gds"
    gdspaths = write_cells(gdspath=gdspath, dirpath="extra/gds", recursively=True)
    assert len(gdspaths) == 10, len(gdspaths)


def test_write_cells():
    gdspath = PATH.gdsdir / "alphabet_3top_cells.gds"
    gdspaths = write_cells(gdspath=gdspath, dirpath="extra/gds", recursively=False)
    assert len(gdspaths) == 3, len(gdspaths)


if __name__ == "__main__":
    test_write_cells_recursively()
    # gdspath = PATH.gdsdir / "alphabet_3top_cells.gds"
    # gdspaths = write_cells(gdspath=gdspath, dirpath="extra/gds", recursively=False)
    # assert len(gdspaths) == 3, len(gdspaths)

    # test_write_cells()
    # import gdsfactory as gf

    # gdspath = PATH.gdsdir / "mzi2x2.gds"
    # gdspaths = write_cells(gdspath=gdspath, dirpath="extra/gds", recursively=False)
    # assert len(gdspaths) == 10, len(gdspaths)

    # gdspath = PATH.gdsdir / "mzi2x2.gds"
    # gf.show(gdspath)
    # gdspaths = write_cells(gdspath=gdspath, dirpath="extra/gds")
    # print(len(gdspaths))

    # sample_pdk_cells = gf.grid(
    #     [
    #         gf.components.straight,
    #         gf.components.bend_euler,
    #         gf.components.grating_coupler_elliptical,
    #     ]
    # )
    # sample_pdk_cells.write_gds("extra/pdk.gds")
    # gf.write_cells.write_cells(gdspath="extra/pdk.gds", dirpath="extra/gds")
    # print(gf.write_cells.get_import_gds_script("extra/gds", module="sky130.components"))



"""Sagnac loop_mirror."""

from __future__ import annotations

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.mmi1x2 import mmi1x2
from gdsfactory.components.spiral_external_io import spiral_external_io
from gdsfactory.routing.manhattan import route_manhattan
from gdsfactory.typings import ComponentSpec


@gf.cell
def loop_mirror(
    component: ComponentSpec = mmi1x2, bend90: ComponentSpec = "bend_euler"
) -> Component:
    """Returns Sagnac loop_mirror.

    Args:
        component: 1x2 splitter.
        bend90: 90 deg bend.

    """
    c = Component()
    component = gf.get_component(component)
    bend90 = gf.get_component(bend90)
    cref = c.add_ref(component)
    routes = route_manhattan(
        cref.ports["o3"],
        cref.ports["o2"],
        straight=gf.components.straight,
        bend=bend90,
    )
    c.add(routes.references)
    c.add_port(name="o1", port=cref.ports["o1"])
    c.absorb(cref)
    return c


@gf.cell
def loop_mirror_with_delay(
    loop_mirror: ComponentSpec = loop_mirror, spiral: ComponentSpec = spiral_external_io
) -> Component:
    """Returns loop_mirror with spiral for delay.

    Args:
        loop_mirror: loop_mirror spec.
        spiral: for delay.

    Notes:
        Delay = 13e-12.
        # delay = length/speed
        # length=delay*speed
        13e-12*3e8/4.2*1e6
    """
    c = Component()
    lm = c << gf.get_component(loop_mirror)
    s = c << spiral_external_io()

    lm.connect("o1", s.ports["o1"])
    return c


if __name__ == "__main__":
    # c = loop_mirror()
    # c = loop_mirror_rotated()
    c = loop_mirror_with_delay()
    c.show(show_ports=True)



import warnings
from typing import List, Optional, Callable

import numpy as np
import scipy.optimize
import shapely.geometry as sg
from gdsfactory.component import Port, ComponentReference, Component
from gdsfactory.path import Path
from gdsfactory.generic_tech.layer_map import LAYER
from gdsfactory.get_netlist import difference_between_angles
from gdsfactory.typings import CrossSectionSpec, Route, ComponentSpec, StepAllAngle
from gdsfactory.typings import STEP_DIRECTIVES_ALL_ANGLE as STEP_DIRECTIVES
from gdsfactory.routing.auto_taper import (
    taper_to_cross_section,
    _get_taper_io_port_names,
)
from gdsfactory.path import extrude


BEND_PATH_FUNCS = {
    # 'euler_bend': euler_path,
}

Connector = Callable[..., List[ComponentReference]]


def get_connector(name: str) -> Connector:
    """
    Gets a connector function by name.

    Args:
        name: the name of the connector function to retrieve

    Returns:
        The specified connector function.
    """
    try:
        connector = CONNECTORS[name]
    except KeyError as e:
        raise KeyError(
            f"{name} is not a valid connector type! Valid types are {list(CONNECTORS.keys())}"
        ) from e
    return connector


def vector_intersection(
    p0, a0, p1, a1, max_distance=100000, raise_error=True
) -> Optional[np.ndarray]:
    """
    Gets the intersection point between two vectors, specified by (point, angle) pairs, (p0, a0) and (p1, a1).

    Args:
        p0: x,y location of vector 0
        a0: angle of vector 0 [degrees]
        p1: x,y location of vector 1
        a1: angle of vector 1 [degrees]
        max_distance: maximum search distance for an intersection [um]
        raise_error: if True, raises an error if no intersection is found. Otherwise, returns None in that case.

    Returns:
        The (x,y) point of intersection, if one is found. Otherwise None.
    """
    a0_rad = np.deg2rad(a0)
    a1_rad = np.deg2rad(a1)
    dx0 = max_distance * np.cos(a0_rad)
    dy0 = max_distance * np.sin(a0_rad)
    p0_far = np.asarray(p0) + [dx0, dy0]
    l0 = sg.LineString([p0, p0_far])

    dx1 = max_distance * np.cos(a1_rad)
    dy1 = max_distance * np.sin(a1_rad)
    p1_far = np.asarray(p1) + [dx1, dy1]
    l1 = sg.LineString([p1, p1_far])

    intersect = l0.intersection(l1)
    if isinstance(intersect, sg.Point):
        return intersect.coords[0]
    if raise_error:
        raise ValueError(
            f"Vectors at {tuple(p0)} and {tuple(p1)} with angles {a0} and {a1} do not intersect!"
        )
    else:
        return None


def _line_intercept(p1, a1, p2, a2):
    if (((a2 - a1) % 180) + 180) % 180 == 0:
        raise ValueError("Lines are parallel!")

    k1 = np.tan(np.deg2rad(a1))
    k2 = np.tan(np.deg2rad(a2))
    x1, y1 = p1
    x2, y2 = p2
    if ((a1 % 180) + 180) % 180 == 90:
        return np.array((x1, k2 * (x1 - x2) + y2))
    elif ((a2 % 180) + 180) % 180 == 90:
        return np.array((x2, k1 * (x2 - x1) + y1))
    else:
        xi = (y1 - y2 - x1 * k1 + x2 * k2) / (k2 - k1)
        yi = k1 * (xi - x1) + y1

        return np.array((xi, yi))


def _get_bend_ports(bend):
    # this is a bit of a hack, but o1 < o2, in0 < out0, hopefully there are no other wacky conventions!
    sorted_port_names = sorted(bend.ports.keys())
    return [bend.ports[n] for n in sorted_port_names]


LOW_LOSS_CROSS_SECTIONS = [
    {"cross_section": "strip", "settings": {"width": 1.0}},
    "strip",
]


def low_loss_connector(
    port1: Port,
    port2: Port,
    prioritized_cross_sections: Optional[List[CrossSectionSpec]] = None,
    **kwargs,
) -> List[ComponentReference]:
    """
    Routes between two ports, using the lowest-loss cross-section which will fit.

    Args:
        port1: the starting port
        port2: the ending port
        prioritized_cross_sections: a list of cross-sections, sorted by preference (starting with most preferred). If None, uses the global variable LOW_LOSS_CROSS_SECTIONS

    Keyword Args:
        kwargs are added for API compatibility, but they are ignored.

    Returns:
        A list of component references comprising the connection
    """
    distance = np.sqrt(np.sum(np.square(port2.center - port1.center)))
    if prioritized_cross_sections is None:
        prioritized_cross_sections = LOW_LOSS_CROSS_SECTIONS
    # try to route with the lowest-loss cross-section
    for low_loss_cs in prioritized_cross_sections:
        taper1 = taper_to_cross_section(port1, cross_section=low_loss_cs)
        taper2 = taper_to_cross_section(port2, cross_section=low_loss_cs)
        taper_lengths = [
            taper.info["length"] for taper in (taper1, taper2) if taper is not None
        ]
        total_taper_length = sum(taper_lengths)
        if total_taper_length < distance:
            refs = []
            if taper1:
                output_port_name = _get_taper_io_port_names(taper1)[1]
                port1 = taper1.ports[output_port_name]
                refs.append(taper1)
            if taper2:
                output_port_name = _get_taper_io_port_names(taper2)[1]
                port2 = taper2.ports[output_port_name]
            intermediate_connector = straight_connector(
                port1, port2, cross_section=low_loss_cs
            )
            refs += intermediate_connector
            if taper2:
                refs.append(taper2)
            return refs
    if port1.cross_section == port2.cross_section:
        # if both cross-sections are the same, keep it
        return straight_connector(port1, port2, cross_section=port1.cross_section)
    elif port1.layer == port2.layer:
        # if the layer is the same, put a width taper, maximizing length of the fatty
        if port2.width > port1.width:
            taper = taper_to_cross_section(port1, port2.cross_section)
            refs = [taper]
            output_port_name = _get_taper_io_port_names(taper)[1]
            refs += straight_connector(
                taper.ports[output_port_name], port2, cross_section=port2.cross_section
            )
        else:
            taper = taper_to_cross_section(port2, port1.cross_section)
            output_port_name = _get_taper_io_port_names(taper1)[1]
            refs = straight_connector(
                port1, taper.ports[output_port_name], cross_section=port2.cross_section
            )
            refs.append(taper)
        return refs
    else:
        # if cross-sections are different, just put the cross-section at the start
        taper = taper_to_cross_section(port1, port2.cross_section)
        refs = [taper]
        output_port_name = _get_taper_io_port_names(taper1)[1]
        refs += straight_connector(
            taper.ports[output_port_name], port2, cross_section=port2.cross_section
        )
        return refs


def _make_error_trace(port1: Port, port2: Port, message: str):
    from gdsfactory.routing.manhattan import RouteWarning

    warnings.warn(message, RouteWarning)
    path = Path([port1.center, port2.center])
    error_component = extrude(path, layer=LAYER.ERROR_PATH, width=1)
    error_ref = ComponentReference(error_component)
    return [error_ref]


def straight_connector(
    port1: Port, port2: Port, cross_section: CrossSectionSpec = "strip"
) -> List[ComponentReference]:
    """
    Connects between the two ports with a straight of the given cross-section.

    Args:
        port1: the starting port
        port2: the ending port
        cross_section: the cross-section to use
    Returns:
        A list of component references comprising the connection
    """
    if np.array_equal(port1.center, port2.center):
        return []
    path = Path([port1.center, port2.center])
    # in usual cases, these angles should be the same, unless they are on opposite sides, in which they are 180 degrees separated
    if abs(difference_between_angles(path.start_angle, port1.orientation)) > 1:
        return _make_error_trace(
            port1,
            port2,
            message=f"Not enough room to route between ports: {port1} and {port2}",
        )

    straight_component = extrude(path, cross_section=cross_section)
    # straight_component = get_component(straight, length=distance, cross_section=cross_section)
    straight_ref = ComponentReference(straight_component)
    # straight_ref.connect('in0', port1)
    return [straight_ref]


def auto_taper_connector(
    port1: Port,
    port2: Port,
    cross_section: CrossSectionSpec = "strip",
    inner_connector: Connector = straight_connector,
) -> List[ComponentReference]:
    """
    Connects the two ports with a straight in the specified cross_section, adding tapers at either end if necessary.

    Args:
        port1: the first port
        port2: the final port
        cross_section: the primary cross section to use for the route
        inner_connector: the connector to use after attaching tapers
    Returns:
        A list of references comprising the connection
    """
    taper1 = taper_to_cross_section(port1, cross_section)
    taper2 = taper_to_cross_section(port2, cross_section)
    route_refs = []
    if taper1:
        route_refs.append(taper1)
        output_port_name = _get_taper_io_port_names(taper1)[1]
        port1 = taper1.ports[output_port_name]
    if taper2:
        route_refs.append(taper2)
        output_port_name = _get_taper_io_port_names(taper2)[1]
        port2 = taper2.ports[output_port_name]
    conn = inner_connector(port1, port2, cross_section)
    route_refs += conn
    return route_refs


CONNECTORS = {
    "low_loss": low_loss_connector,
    "simple": straight_connector,
    "auto_taper": auto_taper_connector,
    None: straight_connector,
}
"""A dictionary of named connectors which can be used for all-angle routing"""


def _place_bend(bend_component: Component, position, rotation) -> ComponentReference:
    """
    Places a bend by its control point at a given position and rotation. The control point of a bend is the intersection of the inverted port vectors.

    Args:
        bend_component: the bend component
        position: the (x,y) position to place the bend
        rotation: the rotation of the bend
    Returns:
        The resulting bend ComponentReference
    """
    bend_ports = _get_bend_ports(bend_component)
    bend_control_point = vector_intersection(
        bend_ports[0].center,
        bend_ports[0].orientation + 180,
        bend_ports[1].center,
        bend_ports[1].orientation + 180,
    )
    bend_ref = ComponentReference(bend_component)
    bend_ref.rotate(
        rotation + 180 - bend_ports[0].orientation, center=bend_control_point
    )
    bend_ref.move(origin=bend_control_point, destination=position)
    return bend_ref


def _point_intersects_ray(p0, a0, p1, angle_tolerance=1e-4):
    x0, y0 = p0
    x1, y1 = p1
    a1 = np.arctan2(y1 - y0, x1 - x0)
    a1 = np.rad2deg(a1)
    return abs(difference_between_angles(a1, a0)) < angle_tolerance


def _null_handler(refs):
    return None


def _all_angle_connector(
    port1: Port,
    port2: Port,
    bend_angle: float,
    intersect: np.ndarray,
    bend: ComponentSpec = "euler_bend",
    cross_section: CrossSectionSpec = "strip",
    connector1: Connector = straight_connector,
    cross_section1: Optional[CrossSectionSpec] = None,
    connector2: Connector = straight_connector,
    cross_section2: Optional[CrossSectionSpec] = None,
    report_segment_separation: Optional[
        Callable[[List[ComponentReference]], None]
    ] = None,
):
    if cross_section1 is None:
        cross_section1 = cross_section
    if cross_section2 is None:
        cross_section2 = cross_section
    if report_segment_separation is None:
        report_segment_separation = _null_handler

    # in the case that the two ports already directly align
    if bend_angle == 0 and _point_intersects_ray(
        port1.center, port1.orientation, port2.center
    ):
        straight_connection = connector2(port1, port2, cross_section=cross_section2)
        report_segment_separation(straight_connection)
        return straight_connection
    if intersect is None:
        # if difference_between_angles(port2.orientation, port1.orientation) == 180:
        sample_bend = _get_bend(bend, angle=90, cross_section=cross_section)
        bend_cs = _get_bend_ports(sample_bend)[0].cross_section
        taper1 = taper_to_cross_section(port1, bend_cs)
        taper2 = taper_to_cross_section(port2, bend_cs)
        route_refs = []
        if taper1:
            route_refs.append(taper1)
            output_port_name = _get_taper_io_port_names(taper1)[1]
            port1 = taper1.ports[output_port_name]
        if taper2:
            route_refs.append(taper2)
            output_port_name = _get_taper_io_port_names(taper2)[1]
            port2 = taper2.ports[output_port_name]
        # try:
        bend_angles = _get_bend_angles(
            port1.center, port2.center, port1.orientation, port2.orientation, bend=bend
        )
        bends = [
            _get_bend(bend, angle=bend_angle, cross_section=cross_section)
            for bend_angle in bend_angles
        ]
        bend_refs = [ComponentReference(b) for b in bends]
        bend_refs_ports = [_get_bend_ports(br) for br in bend_refs]
        bend_refs[0].connect(bend_refs_ports[0][0], port1)
        bend_refs[1].connect(bend_refs_ports[1][0], port2)
        bend_refs_ports = [_get_bend_ports(br) for br in bend_refs]
        connection = connector2(
            bend_refs_ports[0][1], bend_refs_ports[1][1], cross_section=cross_section2
        )
        route_refs += bend_refs + connection
        report_segment_separation(connection + [bend_refs[0]])
        # except Exception as e:
        #     failure_message = f'Unable to complete route! Error message when attempting to create S bend between ports at {port1.center} and {port2.center}: {e}'
        #     route_refs += _make_error_trace(port1, port2, message=failure_message)
        return route_refs
        # return _make_error_trace(port1, port2, f'Port vectors do not intersect: {port1} and {port2}')
    bend_component = _get_bend(bend, angle=bend_angle, cross_section=cross_section)
    bend_ref = _place_bend(
        bend_component, position=intersect, rotation=port1.orientation
    )
    bend_ref_ports = _get_bend_ports(bend_ref)

    straight1 = connector1(port1, bend_ref_ports[0], cross_section=cross_section1)
    straight2 = connector2(bend_ref_ports[1], port2, cross_section=cross_section2)
    route_refs = straight1 + [bend_ref] + straight2
    report_segment_separation(straight1)
    return route_refs


def _get_bend(
    component: ComponentSpec,
    angle: float,
    cross_section: CrossSectionSpec,
    angle_precision: int = 9,
):
    from gdsfactory.pdk import get_component

    if (
        isinstance(component, dict)
        and "settings" in component
        and "cross_section" in component["settings"]
    ):
        return get_component(component, angle=round(angle, angle_precision))
    return get_component(
        component, angle=round(angle, angle_precision), cross_section=cross_section
    )


def _get_bend_angles(p0, p1, a0, a1, bend):
    """get the direct line between the two points."""
    from gdsfactory.pdk import get_component

    a_connect = np.arctan2(p1[1] - p0[1], p1[0] - p0[0])
    a_connect_deg = np.rad2deg(a_connect)
    # these are the angles which should be swept by the bends, if the bends were to take up no space
    bend_angle_ideal_0 = difference_between_angles(a_connect_deg, a0)
    bend_angle_ideal_1 = difference_between_angles(a_connect_deg + 180, a1)
    # retrieves a function to calculate just the bend path (for efficiency) if available
    bend_path_func = BEND_PATH_FUNCS.get(bend)

    def optimization_func(d_angle):
        # apply a delta to both bend angles
        bend_angle0 = bend_angle_ideal_0 + d_angle
        bend_angle1 = bend_angle_ideal_1 + d_angle
        if bend_path_func is None:
            bend0 = get_component(bend, angle=bend_angle0).ref()
            bend1 = get_component(bend, angle=bend_angle1).ref()
            bend0 = bend0.rotate(a0).move(p0)
            bend1 = bend1.rotate(a1).move(p1)
            bend0_output_port = _get_bend_ports(bend0)[1]
            bend1_output_port = _get_bend_ports(bend1)[1]
            dx, dy = bend1_output_port.center - bend0_output_port.center
        else:
            bend0 = Path(bend_path_func(angle=bend_angle0))
            bend1 = Path(bend_path_func(angle=bend_angle1))
            # get rotated coordinates of bends and check if dx/dy match up to tan(angle)
            bend0 = bend0.rotate(a0).move(p0)
            bend1 = bend1.rotate(a1).move(p1)
            dx, dy = bend1.points[-1] - bend0.points[-1]
        angle_est = np.rad2deg(np.arctan2(dy, dx))
        angle_actual = a0 + bend_angle0
        angle_error = abs(difference_between_angles(angle_actual, angle_est))
        return angle_error

    result = scipy.optimize.minimize_scalar(optimization_func, bounds=(-45, 45))
    d_angle = result.x
    bend_angle_0 = bend_angle_ideal_0 + d_angle
    bend_angle_1 = bend_angle_ideal_1 + d_angle
    return bend_angle_0, bend_angle_1


def _get_minimum_separation(refs: List[ComponentReference], *ports) -> float:
    all_ports = [p for ref in refs for p in ref.ports.values()]
    all_ports.extend(ports)
    max_specified_separation = 0
    for port in all_ports:
        if port.cross_section and "separation" in port.cross_section.info:
            separation = port.cross_section.info["separation"]
            if separation > max_specified_separation:
                max_specified_separation = separation

    if max_specified_separation == 0:
        raise ValueError(
            "Cannot automatically determine separation. No ports in route have a cross_section which declares a default separation value!"
        )
    return max_specified_separation


def _points_approx_equal(
    point1: np.ndarray, point2: np.ndarray, tolerance: float = 5e-4
) -> bool:
    return np.sqrt(np.sum(np.square(point1 - point2))) < tolerance


def _angles_approx_opposing(angle1: float, angle2: float, tolerance: float = 1e-4):
    return abs(difference_between_angles(angle1 + 180, angle2)) < tolerance


def get_bundle_all_angle(
    ports1: List[Port],
    ports2: List[Port],
    steps: Optional[List[StepAllAngle]] = None,
    cross_section: CrossSectionSpec = "strip",
    bend: ComponentSpec = "bend_euler",
    connector: str = "low_loss",
    start_angle: Optional[float] = None,
    end_angle: Optional[float] = None,
    end_connector: Optional[str] = None,
    end_cross_section: Optional[CrossSectionSpec] = None,
    separation: Optional[float] = None,
    **kwargs,
) -> List[Route]:
    """Connects a bundle of ports, allowing steps which create waypoints at \
            arbitrary, non-manhattan angles.

    Args:
        ports1: ports at the start of the bundle.
        ports2: ports at the end of the bundle.
        steps: a list of steps, which contain directives on how to proceed with the route.
            "x", "y", "dx", "dy", "ds", "exit_angle", "cross_section", "connector", "separation".
            The first route, between ports1[0] and ports2[0] will take on the role of the primary route,
            and other routes will follow, given the bundling logic.
            It is assume that both ports1 and ports2 are sorted.
        cross_section: cross section of the bends.
            Then the specified connector may also use this information for straights in between.
        bend: the default component to use for the bends.
        connector: the default connector to use to connect between two ports.
        start_angle: if defined and different from the angle of port1,
            will cap the starting port with a bend, as to exit with this angle.
        end_angle:  if defined, and different from the angle of port2,
            will cap the ending port with a bend, as to exit with this angle.
        end_connector: specifies the connector to use for the final straight segment of the route.
        end_cross_section: specifies the cross section to use for the final straight segment of the route.
        separation: specifies the separation between adjacent routes.
            If None, will query each segment's cross-section's and choose the largest value.
        kwargs: added for compatibility, but in general, kwargs will be ignored with a warning.

    Returns:
        List of Routes between ports1 and ports2.

    .. plot::
        :include-source:

        import gdsfactory as gf
        c = gf.Component("demo")

        mmi = gf.components.mmi2x2(width_mmi=10, gap_mmi=3)
        mmi1 = c << mmi
        mmi2 = c << mmi

        mmi2.move((100, 30))
        mmi2.rotate(30)

        routes = gf.routing.get_bundle_all_angle(
            mmi1.get_ports_list(orientation=0),
            [mmi2.ports["o2"], mmi2.ports["o1"]],
            connector=None,
        )
        for route in routes:
            c.add(route.references)
        c.plot()


    """
    from gdsfactory.pdk import get_cross_section

    if kwargs:
        warnings.warn(
            f"Unrecognized arguments for all-angle route will be ignored: {kwargs}"
        )

    connector_func = get_connector(connector)
    routes = []
    is_primary_route = True
    final_connector_func = connector_func
    final_cross_section = cross_section
    waypoints, angles = None, None

    # by default, the second to last segment (in case of a two-step end connection) should have default
    # connector and cross-section
    semi_final_connector_func = connector_func
    semi_final_cross_section = cross_section
    # however, this can be overridden by providing a final step without any directional arguments
    if steps and {"connector", "cross_section"}.issuperset(steps[-1]):
        final_step = steps[-1]
        steps = steps[:-1]
        if "cross_section" in final_step:
            semi_final_cross_section = final_step["cross_section"]
            semi_final_connector_func = auto_taper_connector
        if "connector" in final_step:
            semi_final_connector_func = get_connector(final_step["connector"])

    segment_separations = []

    if end_connector or end_cross_section:
        if end_cross_section:
            final_cross_section = end_cross_section
            final_connector_func = auto_taper_connector
        if end_connector:
            final_connector_func = get_connector(end_connector)

    for port1, port2 in zip(ports1, ports2):
        if _points_approx_equal(port1.center, port2.center) and _angles_approx_opposing(
            port1.orientation, port2.orientation
        ):
            continue
        route_refs = []
        if (
            start_angle is not None
            and difference_between_angles(start_angle, port1.orientation) != 0
        ):
            bend_angle = difference_between_angles(start_angle, port1.orientation)
            bend_component = _get_bend(
                bend, angle=bend_angle, cross_section=cross_section
            )
            bend_ref = ComponentReference(bend_component)
            bend_ref_ports = _get_bend_ports(bend_ref)
            initial_taper = taper_to_cross_section(
                port1, bend_ref_ports[0].cross_section
            )
            if initial_taper:
                route_refs.append(initial_taper)
                output_port_name = _get_taper_io_port_names(initial_taper)[1]
                port1 = initial_taper.ports[output_port_name]
            bend_ref.connect(bend_ref_ports[0], port1)
            bend_ref_ports = _get_bend_ports(bend_ref)
            route_refs.append(bend_ref)
            port1 = bend_ref_ports[1]
        if (
            end_angle is not None
            and difference_between_angles(end_angle, port2.orientation) != 0
        ):
            bend_angle = difference_between_angles(end_angle, port2.orientation)
            bend_component = _get_bend(
                bend, angle=bend_angle, cross_section=cross_section
            )
            bend_ref = ComponentReference(bend_component)
            bend_ref_ports = _get_bend_ports(bend_ref)
            end_taper = taper_to_cross_section(port2, bend_ref_ports[0].cross_section)
            if end_taper:
                route_refs.append(end_taper)
                output_port_name = _get_taper_io_port_names(end_taper)[1]
                port2 = end_taper.ports[output_port_name]
            bend_ref.connect(bend_ref_ports[0], port2)
            bend_ref_ports = _get_bend_ports(bend_ref)
            route_refs.append(bend_ref)
            port2 = bend_ref_ports[1]

        if not is_primary_route and steps:
            # for non-primary routes in the bundle, reset the steps for each new route,
            # based on the primary route's waypoints and angles
            these_waypoints = [port1.center]
            these_angles = [port1.orientation]
            i_step = 0
            intercept_sign = (
                1
                if vector_intersection(
                    these_waypoints[i_step],
                    these_angles[i_step],
                    waypoints[i_step + 1],
                    angles[i_step + 1] + 90,
                    raise_error=False,
                )
                is not None
                else -1
            )
            for i_waypoint in range(1, len(waypoints) - 2):
                # here we need the pitch for the *next* segment, after the bend
                pitch = segment_separations[i_waypoint]
                # the angle orthogonal from the next section's angle of propagation
                offset_angle = angles[i_waypoint] + 90 * intercept_sign

                # offset the next waypoint out by the desired pitch
                offset_pt = waypoints[i_waypoint] + pitch * np.array(
                    [np.cos(np.deg2rad(offset_angle)), np.sin(np.deg2rad(offset_angle))]
                )
                # the next waypoint will be the intersect of the current vector and the line offset from the previous route's next segment
                next_waypoint = _line_intercept(
                    offset_pt,
                    angles[i_waypoint],
                    these_waypoints[i_waypoint - 1],
                    these_angles[i_waypoint - 1],
                )

                these_waypoints.append(next_waypoint)
                these_angles.append(angles[i_waypoint])
            # final_intercept = vector_intersection(these_waypoints[-1], these_angles[-1], port2.center,
            #                                       port2.orientation, raise_error=True)
            waypoints = these_waypoints
            angles = these_angles
            has_explicit_end_angle = True

        if steps and is_primary_route:
            x0, y0 = port1.center
            a0 = port1.orientation
            waypoints = [(x0, y0)]
            angles = [port1.orientation]
            a_final = None

            # for each step, get the next waypoint
            last_step_index = len(steps) - 1
            for i_step, step in enumerate(steps):
                x1, y1 = None, None
                if not STEP_DIRECTIVES.issuperset(step):
                    invalid_step_directives = list(set(step.keys()) - STEP_DIRECTIVES)
                    raise ValueError(
                        f"Invalid step directives: {invalid_step_directives}. Valid directives are {list(STEP_DIRECTIVES)}"
                    )
                if not {"x", "y", "dx", "dy", "ds"}.isdisjoint(step):
                    if "x" in step or "dx" in step:
                        x1 = step.get("x", x0)
                        x1 += step.get("dx", 0)
                    if "y" in step or "dy" in step:
                        y1 = step.get("y", y0)
                        y1 += step.get("dy", 0)
                    if x1 is not None and y1 is not None and a0 is not None:
                        raise ValueError(
                            "Route is overconstrained! x and y and incoming angle are all defined. Please remove one"
                        )
                    if "ds" in step:
                        if {"x", "y", "dx", "dy"}.isdisjoint(step):
                            if a0 is not None:
                                ds = step["ds"]
                                dx = ds * np.cos(np.deg2rad(a0))
                                dy = ds * np.sin(np.deg2rad(a0))
                                x1 = x0 + dx
                                y1 = y0 + dy
                                if i_step == last_step_index:
                                    a_final = a0
                            else:
                                raise ValueError(
                                    'When specifying "ds" as a step, the previous step must have an explicit exit_angle'
                                )
                        else:
                            raise ValueError(
                                f"Route is overconstrained! ds is defined as well as x/y/dx/dy: {step}"
                            )
                    if x1 is None:
                        if a0 is not None:
                            # get intercept with desired y value
                            x1, _ = vector_intersection(
                                (x0, y0), a0, (-1e6, y1), 0, max_distance=2e6
                            )
                            if i_step == last_step_index:
                                a_final = a0
                        elif i_step == last_step_index:
                            if "exit_angle" in step:
                                # if exit_angle is set, assume the segment is vertical, from the previous point to the specified y
                                x1 = x0
                            else:
                                # otherwise, let x be at the intercept of the specified y with the ray defined by port 2's vector
                                x1, _ = vector_intersection(
                                    port2.center,
                                    port2.orientation,
                                    (-1e6, y1),
                                    0,
                                    max_distance=2e6,
                                )
                        else:
                            x1 = x0
                    elif y1 is None:
                        if a0 is not None:
                            # get intercept with desired x value
                            _, y1 = vector_intersection(
                                (x0, y0), a0, (x1, -1e6), 90, max_distance=2e6
                            )
                            if i_step == last_step_index:
                                a_final = a0
                        elif i_step == last_step_index:
                            if "exit_angle" in step:
                                y1 = y0
                            else:
                                _, y1 = vector_intersection(
                                    port2.center,
                                    port2.orientation,
                                    (x1, -1e6),
                                    90,
                                    max_distance=2e6,
                                )
                        else:
                            y1 = y0
                    elif a0 is None and i_step == last_step_index:
                        a_final = np.rad2deg(np.arctan2(y1 - y0, x1 - x0))

                    waypoints.append((x1, y1))
                    a0 = step.get("exit_angle", a_final)
                    angles.append(a0)
                    x0, y0 = x1, y1
                else:
                    raise ValueError(
                        f"Unable to process improperly or incompletely formed step routing command: {step}"
                    )
            has_explicit_end_angle = angles[-1] is not None

        if steps:
            waypoints.append(port2.center)
            bends = []
            prev_port = port1
            # go back over the waypoints and calculate unspecified angles and bends
            for i_step, step in enumerate(steps):
                # override the current cross section and connector, if applicable
                override_cs = steps[i_step].get("cross_section")
                override_connector = steps[i_step].get("connector")
                # override the current separation, if applicable
                override_separation = step.get("separation")
                if override_cs:
                    this_cs = get_cross_section(override_cs)
                    this_connector = auto_taper_connector
                else:
                    this_cs = cross_section
                    this_connector = connector_func
                if override_connector:
                    this_connector = get_connector(override_connector)

                # build up the sequence of references in the current segment of the route
                if i_step + 1 < len(angles):
                    # the angle going into the current step
                    # angle 0 should always be explicitly defined, from the first port
                    angle0 = angles[i_step]
                    # the angle going out of the current step
                    angle_next = angles[i_step + 1]
                    # if no angle was explicitly defined, let's calculate it here
                    if angle_next is None:
                        p1 = waypoints[i_step + 2]
                        p0 = waypoints[i_step + 1]
                        angle_next = np.rad2deg(
                            np.arctan2(p1[1] - p0[1], p1[0] - p0[0])
                        )
                    dangle = difference_between_angles(angle_next, angle0)
                    if dangle == 180:
                        raise ValueError(
                            "Intermediate 180 degree bends are not currently supported!"
                        )
                    elif dangle == 0:
                        next_port = prev_port.flip()
                        next_port.center = waypoints[i_step + 1]
                        connection = this_connector(
                            prev_port, next_port, cross_section=this_cs
                        )
                        route_refs += connection
                        prev_port = next_port.flip()
                    else:
                        bend_component = _get_bend(
                            bend, angle=dangle, cross_section=cross_section
                        )
                        bend_ref = _place_bend(
                            bend_component,
                            position=waypoints[i_step + 1],
                            rotation=angle0,
                        )
                        bend_ref_ports = _get_bend_ports(bend_ref)
                        connection = this_connector(
                            prev_port, bend_ref_ports[0], cross_section=this_cs
                        )
                        route_refs += connection
                        route_refs.append(bend_ref)
                        bends.append(bend_ref)
                        prev_port = bend_ref_ports[1]
                    angles[i_step + 1] = angle_next
                    if is_primary_route:
                        if override_separation is not None:
                            this_separation = override_separation
                        elif separation is not None:
                            this_separation = separation
                        else:
                            this_separation = _get_minimum_separation(
                                connection, prev_port
                            )
                        segment_separations.append(this_separation)

            if has_explicit_end_angle:
                port1 = prev_port
            else:
                if _point_intersects_ray(
                    port2.center, port2.orientation, prev_port.center
                ):
                    final_connection = final_connector_func(
                        prev_port, port2, cross_section=final_cross_section
                    )
                    route_refs += final_connection
                    this_separation = _get_minimum_separation(
                        final_connection, prev_port
                    )
                    segment_separations.append(this_separation)
                else:
                    route_refs += _make_error_trace(
                        prev_port,
                        port2,
                        "Cannot complete final step of route! "
                        "Try setting an exit_angle in your final "
                        "step which intersects the vector of the destination port.",
                    )

        if not steps or has_explicit_end_angle:
            intersect = vector_intersection(
                port1.center,
                port1.orientation,
                port2.center,
                port2.orientation,
                raise_error=False,
            )
            report_segment_separation = None

            def _report_separations_w_steps(refs):
                segment_separations.append(_get_minimum_separation(refs))

            if steps:
                angles.insert(-1, port1.orientation)
                waypoints.insert(-1, intersect)
                report_segment_separation = _report_separations_w_steps
            bend_angle = difference_between_angles(
                port2.orientation + 180, port1.orientation
            )
            final_connection = _all_angle_connector(
                port1,
                port2,
                bend_angle=bend_angle,
                intersect=intersect,
                bend=bend,
                cross_section=cross_section,
                connector1=semi_final_connector_func,
                connector2=final_connector_func,
                cross_section1=semi_final_cross_section,
                cross_section2=final_cross_section,
                report_segment_separation=report_segment_separation,
            )
            route_refs += final_connection
            this_separation = _get_minimum_separation(final_connection, port1)
            segment_separations.append(this_separation)
        route_length = sum(r.info["length"] for r in route_refs)
        route = Route(references=route_refs, ports=(port1, port2), length=route_length)
        routes.append(route)
        is_primary_route = False
    return routes


if __name__ == "__main__":
    import gdsfactory as gf

    c = gf.Component("demo")

    mmi = gf.components.mmi2x2(width_mmi=10, gap_mmi=3)
    mmi1 = c << mmi
    mmi2 = c << mmi

    mmi2.move((100, 30))
    mmi2.rotate(30)

    routes = gf.routing.get_bundle_all_angle(
        mmi1.get_ports_list(orientation=0),
        [mmi2.ports["o2"], mmi2.ports["o1"]],
        connector=None,
    )
    for route in routes:
        c.add(route.references)
    c.show()



"""Based on phidl.geometry."""
from __future__ import annotations

from typing import Tuple, Union

import gdstk

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.component_layout import Polygon, _parse_layer
from gdsfactory.component_reference import ComponentReference
from gdsfactory.typings import ComponentOrReference, Int2, LayerSpec


@gf.cell
def boolean(
    A: Union[ComponentOrReference, Tuple[ComponentOrReference, ...]],
    B: Union[ComponentOrReference, Tuple[ComponentOrReference, ...]],
    operation: str,
    precision: float = 1e-4,
    num_divisions: Union[int, Int2] = (1, 1),
    layer: LayerSpec = (1, 0),
) -> Component:
    """Performs boolean operations between 2 Component/Reference/list objects.

    ``operation`` should be one of {'not', 'and', 'or', 'xor', 'A-B', 'B-A', 'A+B'}.
    Note that 'A+B' is equivalent to 'or', 'A-B' is equivalent to 'not', and
    'B-A' is equivalent to 'not' with the operands switched

    You can also use gdsfactory.drc.boolean_klayout

    Args:
        A: Component(/Reference) or list of Component(/References).
        B: Component(/Reference) or list of Component(/References).
        operation: {'not', 'and', 'or', 'xor', 'A-B', 'B-A', 'A+B'}.
        precision: float Desired precision for rounding vertex coordinates.
        num_divisions: number of divisions with which the geometry is divided into
          multiple rectangular regions. This allows for each region to be
          processed sequentially, which is more computationally efficient.
        layer: Specific layer to put polygon geometry on.

    Returns: Component with polygon(s) of the boolean operations between
      the 2 input Components performed.

    Notes
    -----
    'A+B' is equivalent to 'or'.
    'A-B' is equivalent to 'not'.
    'B-A' is equivalent to 'not' with the operands switched.

    """
    D = Component()
    A_polys = []
    B_polys = []
    A = list(A) if isinstance(A, (list, tuple)) else [A]
    B = list(B) if isinstance(B, (list, tuple)) else [B]

    for X, polys in ((A, A_polys), (B, B_polys)):
        for e in X:
            if isinstance(e, (Component, ComponentReference)):
                polys.extend(e.get_polygons())
            elif isinstance(e, Polygon):
                polys.extend(e.polygons)

    layer = gf.pdk.get_layer(layer)
    gds_layer, gds_datatype = _parse_layer(layer)

    operation = operation.lower().replace(" ", "")
    if operation == "a-b":
        operation = "not"
    elif operation == "b-a":
        operation = "not"
        A_polys, B_polys = B_polys, A_polys
    elif operation == "a+b":
        operation = "or"
    elif operation not in ["not", "and", "or", "xor", "a-b", "b-a", "a+b"]:
        raise ValueError(
            "gdsfactory.geometry.boolean() `operation` "
            "parameter not recognized, must be one of the "
            "following:  'not', 'and', 'or', 'xor', 'A-B', "
            "'B-A', 'A+B'"
        )

    # Check for trivial solutions
    if (not A_polys or not B_polys) and operation != "or":
        if (
            operation != "not"
            and operation != "and"
            and operation == "xor"
            and not A_polys
            and not B_polys
            or operation != "not"
            and operation == "and"
        ):
            p = None
        elif operation != "not" and operation == "xor" and not A_polys:
            p = B_polys
        elif operation != "not" and operation == "xor":
            p = A_polys
        elif operation == "not":
            p = A_polys or None
    elif not A_polys and not B_polys:
        p = None
    else:
        p = gdstk.boolean(
            operand1=A_polys,
            operand2=B_polys,
            operation=operation,
            precision=precision,
            layer=gds_layer,
            datatype=gds_datatype,
        )

    if p is not None:
        polygons = D.add_polygon(p, layer=layer)
        [polygon.fracture(precision=precision) for polygon in polygons]
    return D


def test_boolean() -> None:
    c = gf.Component()
    e1 = c << gf.components.ellipse()
    e2 = c << gf.components.ellipse(radii=(10, 6))
    e3 = c << gf.components.ellipse(radii=(10, 4))
    e3.movex(5)
    e2.movex(2)
    c = boolean(A=[e1, e3], B=e2, operation="A-B")
    assert len(c.polygons) == 2, len(c.polygons)


if __name__ == "__main__":
    # c = gf.Component()
    # e1 = c << gf.components.ellipse()
    # e2 = c << gf.components.ellipse(radii=(10, 6))
    # e3 = c << gf.components.ellipse(radii=(10, 4))
    # e3.movex(5)
    # e2.movex(2)
    # c = boolean(A=[e1, e3], B=e2, operation="A-B")

    import time

    n = 50
    c1 = gf.c.array(gf.c.circle(radius=10), columns=n, rows=n)
    c2 = gf.c.array(gf.c.circle(radius=9), columns=n, rows=n).movex(5)

    t0 = time.time()
    c = boolean(c1, c2, operation="xor")
    t1 = time.time()
    print(t1 - t0)

    c.show(show_ports=True)



import pathlib
from typing import Optional, Tuple
import time

import numpy as np
from femwell import mode_solver

import gdsfactory as gf
from gdsfactory.config import logger
from gdsfactory.pdk import _ACTIVE_PDK, get_layer_stack
from gdsfactory.simulation.get_modes_path import get_modes_path_femwell
from gdsfactory.technology import LayerStack
from gdsfactory.typings import CrossSectionSpec, PathType, ComponentSpec

from skfem import (
    Basis,
    ElementTriN1,
    ElementTriN2,
    ElementTriP0,
    ElementTriP1,
    ElementTriP2,
    Mesh,
)


def load_mesh_basis(mesh_filename: PathType):
    mesh = Mesh.load(mesh_filename)
    basis = Basis(mesh, ElementTriN2() * ElementTriP2())
    basis0 = basis.with_element(ElementTriP0())
    return mesh, basis0


def compute_cross_section_modes(
    cross_section: CrossSectionSpec,
    layerstack: LayerStack,
    wl: float = 1.55,
    num_modes: int = 4,
    order: int = 1,
    radius: float = np.inf,
    mesh_filename: str = "mesh.msh",
    dirpath: Optional[PathType] = None,
    filepath: Optional[PathType] = None,
    overwrite: bool = False,
    with_cache: bool = True,
    wafer_padding: float = 2.0,
    **kwargs,
):
    """Calculate effective index of a cross-section.

    Defines a "straight" component of the cross_section, and calls compute_component_slice_modes.
    """
    # Get meshable component from cross-section
    c = gf.components.straight(length=10, cross_section=cross_section)
    bounds = c.bbox
    dx = np.diff(bounds[:, 0])[0]

    xsection_bounds = [
        [dx / 2, bounds[0, 1] - wafer_padding],
        [dx / 2, bounds[1, 1] + wafer_padding],
    ]

    # Mesh as component
    return compute_component_slice_modes(
        component=c,
        xsection_bounds=xsection_bounds,
        layerstack=layerstack,
        wl=wl,
        num_modes=num_modes,
        order=order,
        radius=radius,
        mesh_filename=mesh_filename,
        dirpath=dirpath,
        filepath=filepath,
        overwrite=overwrite,
        with_cache=with_cache,
        wafer_padding=wafer_padding,
        **kwargs,
    )


def compute_component_slice_modes(
    component: ComponentSpec,
    xsection_bounds: Tuple[Tuple[float, float], Tuple[float, float]],
    layerstack: LayerStack,
    wl: float = 1.55,
    num_modes: int = 4,
    order: int = 1,
    radius: float = np.inf,
    mesh_filename: str = "mesh.msh",
    dirpath: Optional[PathType] = None,
    filepath: Optional[PathType] = None,
    overwrite: bool = False,
    with_cache: bool = True,
    wafer_padding: float = 2.0,
    **kwargs,
):
    """Calculate effective index of component slice.

    Args:
        component: gdsfactory component.
        xsection_bounds: xy line defining where to take component cross_section.
        layerstack: gdsfactory layerstack.
        wavelength: wavelength (um).
        num_modes: number of modes to return.
        order: order of the mesh elements. 1: linear, 2: quadratic.
        radius: bend radius of the cross-section.
        mesh_filename (str, path): where to save the .msh file. If with_cache, will be filepath.msh.
        dirpath: Optional directory to store modes.
        filepath: Optional path to store modes.
        overwrite: Overwrite mode filepath if it exists.
        with_cache: write modes to filepath cache.
        wafer_padding: padding beyond bbox to add to WAFER layers.
        bend_radius: bend radius for mode solving (typically called from cross_section)

    Keyword Args:
        resolutions (Dict): Pairs {"layername": {"resolution": float, "distance": "float}}
            to roughly control mesh refinement within and away from entity, respectively.
        mesh_scaling_factor (float): factor multiply mesh geometry by.
        default_resolution_min (float): gmsh minimal edge length.
        default_resolution_max (float): gmsh maximal edge length.
        background_tag (str): name of the background layer to add (default: no background added).
        background_padding (Tuple): [xleft, ydown, xright, yup] distances to add to the components and to fill with background_tag.
        global_meshsize_array: np array [x,y,z,lc] to parametrize the mesh.
        global_meshsize_interpolant_func: interpolating function for global_meshsize_array.
        extra_shapes_dict: Optional[OrderedDict] = OrderedDict of {key: geo} with key a label and geo a shapely (Multi)Polygon or (Multi)LineString of extra shapes to override component.
        merge_by_material: boolean, if True will merge polygons from layers with the same layer.material. Physical keys will be material in this case.
    """
    sim_settings = dict(
        xsection_bounds=xsection_bounds,
        wl=wl,
        num_modes=num_modes,
        radius=radius,
        order=order,
        wafer_padding=wafer_padding,
        **kwargs,
    )
    filepath = filepath or get_modes_path_femwell(
        component=component,
        dirpath=dirpath,
        layerstack=layerstack,
        **sim_settings,
    )
    filepath = pathlib.Path(filepath)
    mesh_filename = filepath.with_suffix(".msh") if with_cache else mesh_filename

    if with_cache and filepath.exists():
        if overwrite:
            filepath.unlink()

        else:
            logger.info(f"Simulation loaded from {filepath!r}")

            modes_dict = dict(np.load(filepath))
            mesh, basis0 = load_mesh_basis(mesh_filename)

            if order == 1:
                element = ElementTriN1() * ElementTriP1()
            elif order == 2:
                element = ElementTriN2() * ElementTriP2()
            else:
                raise AssertionError("Only order 1 and 2 implemented by now.")

            basis = basis0.with_element(element)

            return modes_dict["lams"], basis, modes_dict["xs"]

    # Mesh
    mesh = component.to_gmsh(
        type="uz",
        xsection_bounds=xsection_bounds,
        layer_stack=layerstack,
        filename=mesh_filename,
        wafer_padding=wafer_padding,
        **kwargs,
    )

    # Assign materials to mesh elements
    mesh, basis0 = load_mesh_basis(mesh_filename)
    epsilon = basis0.zeros(dtype=complex)
    for layername, layer in layerstack.layers.items():
        if layername in mesh.subdomains.keys():
            epsilon[basis0.get_dofs(elements=layername)] = (
                _ACTIVE_PDK.materials_index[layer.material](wl) ** 2
            )
        if "background_tag" in kwargs:
            epsilon[basis0.get_dofs(elements=kwargs["background_tag"])] = (
                _ACTIVE_PDK.materials_index[kwargs["background_tag"]](wl) ** 2
            )

    # Mode solve
    lams, basis, xs = mode_solver.compute_modes(
        basis0,
        epsilon,
        wavelength=wl,
        mu_r=1,
        num_modes=num_modes,
        order=order,
        radius=radius,
        solver="slepc",
    )

    if with_cache:
        modes_dict = {"lams": lams, "xs": xs}
        np.savez_compressed(filepath, **modes_dict)
        logger.info(f"Write mode to {filepath!r}")
    return lams, basis, xs


if __name__ == "__main__":
    start = time.time()

    filtered_layerstack = LayerStack(
        layers={
            k: get_layer_stack().layers[k]
            for k in (
                "core",
                "clad",
                "slab90",
                "box",
            )
        }
    )

    filtered_layerstack.layers["core"].thickness = 0.2

    resolutions = {
        "core": {"resolution": 0.02, "distance": 2},
        "clad": {"resolution": 0.2, "distance": 1},
        "box": {"resolution": 0.2, "distance": 1},
        "slab90": {"resolution": 0.05, "distance": 1},
    }

    cross_section = False

    if cross_section:
        lams, basis, xs = compute_cross_section_modes(
            cross_section="rib",
            layerstack=filtered_layerstack,
            wl=1.55,
            num_modes=4,
            order=1,
            radius=np.inf,
            mesh_filename="mesh.msh",
            resolutions=resolutions,
            overwrite=True,
            with_cache=True,
        )
        mode_solver.plot_mode(
            basis=basis,
            mode=np.real(xs[0]),
            plot_vectors=False,
            colorbar=True,
            title="E",
            direction="y",
        )
        end = time.time()
        print(end - start)
    else:
        component = gf.components.coupler_full(dw=0)
        component.show()

        compute_component_slice_modes(
            component,
            [[0, -3], [0, 3]],
            layerstack=filtered_layerstack,
            wl=1.55,
            num_modes=4,
            order=1,
            radius=np.inf,
            mesh_filename="./mesh.msh",
            resolutions=resolutions,
            overwrite=True,
            with_cache=True,
        )
